#!/usr/bin/env python

import sys
import os
import re
import glob
import shutil
import tempfile
import argparse
import subprocess

def stop_err( msg ):
    sys.stderr.write( '%s\n' % msg )
    sys.exit()
    
def timenow():
    """return current time as a string
    """
    return time.strftime('%d/%m/%Y %H:%M:%S', time.localtime(time.time()))


class CGATBase():
    """
    simple base class with some utilities for Picard
    adapted and merged with Kelly Vincent's code april 2011 Ross
    lots of changes...
    """
    
    def __init__(self, opts=None,arg0=None):
        """ common stuff needed at init for a picard tool
        """
        pass

    def baseName(self,name=None):
        return os.path.splitext(os.path.basename(name))[0]

    def setLogging(self,logfname="picard_wrapper.log"):
        """setup a logger
        """
        logging.basicConfig(level=logging.INFO,
                    filename=logfname,
                    filemode='a')


    def readLarge(self,fname=None):
        """ read a potentially huge file.
        """
        try:
            # get stderr, allowing for case where it's very large
            tmp = open( fname, 'rb' )
            s = ''
            buffsize = 1048576
            try:
                while True:
                    more = tmp.read( buffsize )
                    if len(more) > 0:
                        s += more
                    else:
                        break
            except OverflowError:
                pass
            tmp.close()
        except Exception, e:
            stop_err( 'Read Large Exception : %s' % str( e ) )   
        return s
    
    def runStatement(self,cl=None,output_dir=None):
        """ construct and run a command line
        we have galaxy's temp path as opt.temp_dir so don't really need isolation
        sometimes stdout is needed as the output - ugly hacks to deal with potentially vast artifacts
        """
        assert cl <> None, 'PicardBase runCL needs a command line as cl'
        process = subprocess.Popen(cl, shell=True )
        rval = process.wait()
    
    def runPic(self, jar, cl):
        """
        cl should be everything after the jar file name in the command
        """
        runme = ['java -Xmx%s' % self.opts.maxjheap]
        runme.append(" -Djava.io.tmpdir='%s' " % self.opts.tmpdir)
        runme.append('-jar %s' % jar)
        runme += cl
        s,stdouts,rval = self.runCL(cl=runme, output_dir=self.opts.outdir)
        return stdouts,rval

    def samToBam(self,infile=None,outdir=None):
        """
        use samtools view to convert sam to bam
        """
        fd,tempbam = tempfile.mkstemp(dir=outdir,suffix='rgutilsTemp.bam')
        cl = ['samtools view -h -b -S -o ',tempbam,infile]
        tlog,stdouts,rval = self.runCL(cl,outdir)
        return tlog,tempbam,rval

    def sortSam(self, infile=None,outfile=None,outdir=None):
        """
        """
        print '## sortSam got infile=%s,outfile=%s,outdir=%s' % (infile,outfile,outdir)
        cl = ['samtools sort',infile,outfile]
        tlog,stdouts,rval = self.runCL(cl,outdir)
        return tlog

    def cleanup(self):
        for fname in self.delme:
            try:
                os.unlink(fname)
            except:
                pass

def __main__():

    # use argparse to ignore unknown options
    parser = argparse.ArgumentParser()

    parser.add_argument( "--version", action="version", version="%(prog)s")
    parser.add_argument( "--wrapper-command", dest="command", type=str)
    parser.add_argument( "--wrapper-bam-file", dest="bam_file", type=str)
    parser.add_argument( "--wrapper-bam-option", dest="bam_option", type=str)
    parser.add_argument( "--wrapper-bai-file", dest="bai_file", type=str)
    parser.add_argument( "--wrapper-dry-run", dest="dry_run", action="store_true")
    parser.add_argument( "--wrapper-html-dir", dest="html_dir", type=str)
    parser.add_argument( "--wrapper-html-file", dest="html_file", type=str)
        
    options, unknown = parser.parse_known_args( )

    cgat = CGATBase( options )

    option_map = []

    if options.bai_file or options.bam_file:
        if not (options.bai_file and options.bam_file):
            raise ValueError("wrapper called with bam or bai file, but not both")
        
        if not options.bam_option:
            options.bam_option = "bam-file"

        tmp_fd, tmp_name = tempfile.mkstemp()
        tmp_bam_name = '%s.bam' % tmp_name
        tmp_bai_name = '%s.bai' % tmp_bam_name
        os.symlink( options.bam_file, tmp_bam_name )
        os.symlink( options.bai_file, tmp_bai_name )
        if options.bam_option.startswith("--"):
            # long option
            option_map.append( "%s=%s" % (options.bam_option, tmp_bam_name) )
        else:
            # short option
            option_map.append( "%s %s" % (options.bam_option, tmp_bam_name) )

    if options.html_dir:
        os.mkdir( options.html_dir )
        option_map.append( "%s=%s/%%s" % ( "--output-filename-pattern", options.html_dir ) )

    statement = "python " + " ".join( [options.command] + unknown + option_map )

    if options.dry_run:
        sys.stdout.write( statement + "\n" )
        return

    else:
        cgat.runStatement( statement )
    
    if options.bai_file:
        os.unlink( tmp_bam_name )
        os.unlink( tmp_bai_name )
    
    if options.html_file:
        with open( options.html_file, "w" ) as outf:
            outf.write('<h1>%s - Output</h1>' % os.path.basename( options.wrapper_command))
            for fn in glob.glob( os.path.join( options.html_dir, "*.*" ) ):
                dirname, basename = os.path.split(fn)
                outf.write('''<li><a href="%s">%s</a></li>\n''' % (basename, basename))
                
if __name__=="__main__": __main__()




#! /bin/env python
################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2011 David Sims
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
vcfstats_sqlite.py - reformat output of vcf-stats for database loading
======================================================================

:Author: David Sims
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

create a csv separated file for loading into a database from 
output of vcf-stats utility in vcf-tools package.

Usage
-----

Example::

   python vcfstats_sqlite.py [files] > [outfile]

Type::

   python vcfstats_sqlite.py --help

for command line help.


Code
----

'''
import os
import sys
import string
import re
import time
import optparse
import tempfile
import subprocess
import types
import CGAT.Experiment as E
import csv
import CGAT.CSV as CSV
import CGAT.IOTools as IOTools

if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: vcfstats_sqlite.py 0001 2011-04-13 davids $", usage = globals()["__doc__"])

    (options, args) = E.Start( parser )

    options.filenames = args

    if len(options.filenames) < 1:
        options.stdout.write("# Error: no vcf-stats files specified/found.")
        sys.exit(1)

    E.info( "Parsing %i file(s)" % len(options.filenames) )

    # set up output files
    vcf_file = open('vcfstats.txt', 'w')
    indel_file = open('indelstats.txt', 'w')
    snp_file = open('snpstats.txt', 'w')
    shared_file = open('sharedstats.txt', 'w')

    for fileno,filename in enumerate(options.filenames):

        prefix = os.path.basename( filename )
        trackname = prefix.replace(".vcfstats","")

        if os.path.exists(filename):
            lines = [ x for x in IOTools.openFile(filename, "r").readlines()]
        else:
            lines = []

        if len(lines) == 0:
            options.stdout.write("# Error: empty vcf-stats file found: $(filename)s")
            sys.exit(1)
        else:
            E.info( "File %i contains %i lines" % (fileno,len(lines)) )
            vcf_stats = dict(track=trackname)
            snp_stats = dict(track=trackname)
            indel_stats = dict()
            shared_stats = dict() 
            all_vars = False
            indels = False
            snps = False
            shared = False 
            for i, line in enumerate(lines):
                line = line.strip()
                if line.find("'all'") > -1: 
                    all_vars = True
                    E.info("Found 'all'")
                    continue

                if all_vars:
                    if line.find("=>") > -1: 
                        fields = line.split("=>")
                        key = fields[0].strip().replace("'","").replace(">","_")
                        val = fields[1].strip().replace(",","")
                    else: 
                        key = "NA"
                        val = "NA"
                    if key=="indel" and val=="{":
                        indels = True
                        E.info("Found 'indels'")
                        continue
                    elif key=="snp" and val=="{":
                        snps = True
                        E.info("Found 'SNPs'")
                        continue
                    elif key=="shared" and val=="{":
                        shared = True
                        E.info("Found 'Shared'")
                        continue

                    if indels:
                        if line.find("}") > -1:
                           indels = False
                           E.info("Processed 'indels'")
                           continue
                        else:
                            indel_stats[key] = val
                    elif snps:
                        if line.find("}") > -1:
                            snps = False
                            E.info("Processed 'SNPs'")
                            continue
                        else:
                            snp_stats[key] = val
                    elif shared:
                        if line.find("}") > -1:
                            shared = False
                            E.info("Processed 'Shared'")
                            continue
                        else:
                            shared_stats[key] = val
                    elif key != "NA":
                        vcf_stats[key] = val 
            
            # Ensure all keys are present
            allkeys = ["nalt_1","nalt_2","nalt_3","nalt_4","nalt_5","track","count","snp_count","indel_count"]
            for k in allkeys:
                if vcf_stats.has_key(k):
                    continue
                else:
                    vcf_stats[k] = "0"

            # Write header (for first file only)
            if filename == options.filenames[0]:

                # Ensure keys are sorted
                srt = vcf_stats.keys()
                srt.sort()
                sep=""
                for k in srt:
                    vcf_file.write("%s%s" % (sep,k))
                    sep="\t"
                vcf_file.write("\n")

                indel_file.write("track\tindel_length\tindel_count\n")
                shared_file.write("track\tno_samples\tvar_count\n")

                sep=""
                for k in snp_stats.iterkeys():
                    snp_file.write("%s%s" % (sep,k))
                    sep="\t"
                snp_file.write("\n")

            # Write data
            sep=""
            srt = vcf_stats.keys()
            srt.sort()
            for k in srt:
                vcf_file.write("%s%s" % (sep,vcf_stats[k]))
                sep="\t"
            vcf_file.write("\n")


            # Check all indel lengths are covered
            r = range(-20,20,1)
            for i in r:
                if indel_stats.has_key(str(i)):
                    continue
                else:
                    indel_stats[i]="0"
            for k in indel_stats.iterkeys():
                indel_file.write("%s\t%s\t%s\n" % (trackname,k,indel_stats[k]))
 
            for k in shared_stats.iterkeys():
                shared_file.write("%s\t%s\t%s\n" % (trackname,k,shared_stats[k]))

            sep=""
            for k in snp_stats.iterkeys():
                snp_file.write("%s%s" % (sep,snp_stats[k]))
                sep="\t"
            snp_file.write("\n")

    # close files
    vcf_file.close()
    indel_file.close()
    snp_file.close()

    E.Stop()
    sys.exit(0)




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
WrapperPhylip.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Code
----

'''
import os, sys, string, re, tempfile, subprocess, optparse, shutil

from types import *

"""Wrapper for CodeML
"""

import Experiment
import TreeTools, MatlabTools

class Error(Exception):
    """Base class for exceptions in this module."""
    pass

    def __str__(self):
        return str(self.message)

class ParsingError(Error):
    """Exception raised for errors while parsing

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message, line):
        self.message = message + " at line " + line

class UsageError(Error):
    """Exception raised for errors while starting

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class PhylipResult:
    def __init__(self):
        self.mNexus = None
        self.mOutfile = None
        self.mMatrix = None

    def parseContrasts( self, infile ):
        """parse results from contrasts run."""
        ## remove empty lines and strip
        
        lines = filter( lambda x: x, map( lambda x: x.strip(), infile.readlines()))

        sections = filter( lambda x: lines[x+1][:3] == "---", range(0,len(lines)-1))
        sections.append(len(lines))
        
        def readTable( lines ):
            data = []
            for line in lines:
                data.append( map( float, re.split( "\s+", line[:-1] ) ) )
            return data

        for s in range(len(sections)-1):
            start = sections[s]
            end = sections[s+1]

            header = lines[start]
            if header[:len("Contrasts")] == "Contrasts":
                self.mContrasts = readTable( lines[start+2:end] )
            elif header[:len("Covariance")] == "Covariance":
                self.mCovariance = readTable( lines[start+2:end] )
            elif header[:len("Correlations")] == "Correlations":
                self.mCorrelations = readTable( lines[start+2:end] )
            elif header[:len("Regressions")] == "Regressions":
                self.mRegressions = readTable( lines[start+2:end] )
                
class Phylip:

    def __init__( self ):
        
        self.mMapInput2Phylip = {}
        self.mMapPhylip2Input = {}
        
        self.mTempdir = None
        
        self.mInputTree = None
        self.mInputTrees = None
        self.mInputData = None
        self.mInputMatrix = None
        self.mInputMali = None
        
        self.mOutputTree = "-"
        
        self.mOptions = ""

        self.mOutputStdout = None
        self.mOutputStderr = None

        self.mPruneTree = False

        self.mLogLevel = 0

    def __del__(self):

        if self.mTempdir and self.mTempdir != "tmp" and self.mLogLevel == 0:
            os.system( "rm -rf %s" % self.mTempdir )

    def setLogLevel( self, loglevel ):
        self.mLogLevel = loglevel 

    def setProgram( self, program ):
        self.mProgram = program

    def setOptions( self, options ):
        self.mOptions = options
        
    def setTree( self, tree ):
        self.mInputTree = tree

    def setTrees( self, trees ):
        self.mInputTrees = trees

    def setData( self, data ):
        """set input data. This is a matrix with row headers."""
        self.mInputData = data

    def setMatrix( self, matrix):
        self.mInputMatrix = matrix

    def setMali( self, mali):
        self.mInputMali = mali

    def setPruneTree( self, flag = True ):
        self.mPruneTree = flag

    def updateMaps( self, taxa ):
        """update maps."""
        for taxon in taxa:
            if taxon not in self.mMapInput2Phylip:
                key = "t_%05i" % (len(self.mMapInput2Phylip) )
                self.mMapInput2Phylip[taxon] = key
                self.mMapPhylip2Input[key] = taxon


    def __reset(self):
        """clear internal state variables."""
        self.mMapInput2Phylip = {}
        self.mMapPhylip2Input = {}
        
        self.mTempdir = None
        
    def prepareRun(self):

        self.__reset()

        self.mTempdir = tempfile.mkdtemp()
        # self.mTempdir = "tmp"
        if not os.path.exists( self.mTempdir ):
            os.mkdir(self.mTempdir)

        if self.mInputMatrix and self.mInputData:
            raise ValueError("please specify either input matrix or input data, but not both.")
        
        ## prepare input matrix. Should already be in phylip like
        ## format, but long identifiers are shortened and tabs are
        ## replaced by spaces.
        if self.mInputMatrix:

            outfile = open(self.mTempdir + "/infile", "w")
            
            identifiers = map( lambda x: re.split("\s+", x[:-1])[0], self.mInputMatrix[1:])
            self.updateMaps( identifiers )

            outfile.write(self.mInputMatrix[0])
            for line in self.mInputMatrix[1:]:
                data = re.split("\s+", line[:-1])
                new_line = self.mMapInput2Phylip[data[0]] + "       " + "  ".join(data[1:])
                outfile.write(new_line + "\n" )
                
            outfile.close()

            if self.mLogLevel >= 1:
                print "# written input matrix with %i taxa to %s" % (len(identifiers), self.mTempdir + "/infile" )
                os.system("cat %s" %  self.mTempdir + "/infile" )

        elif self.mInputData:
            
            outfile = open(self.mTempdir + "/infile", "w")
            outfile.write( "%i %i\n" % (len(self.mInputData), len(self.mInputData[0]) -1 ) )
            identifiers = map( lambda x: x[0], self.mInputData) 
            self.updateMaps( identifiers)
            
            for x in range(len(identifiers)):
                outfile.write( "%-10s %s\n" % (self.mMapInput2Phylip[identifiers[x]], " ".join(self.mInputData[x][1:]) ) )

            outfile.close()
            
            if self.mLogLevel >= 1:
                print "# written input matrix with %i taxa to %s" % (len(identifiers), self.mTempdir + "/infile" )
                os.system("cat %s" %  self.mTempdir + "/infile" )
                
        ## prepare input tree or trees
        self.mNInputTrees = 0
        if self.mInputTree or self.mInputTrees:

            outfile = open(self.mTempdir + "/intree", "w")

            if self.mInputTree and self.mInputTrees:
                raise UsageError("please supply either one or mupltiple trees, but not both." )

            if self.mInputTree:
                trees = [ self.mInputTree ]
            else:
                trees = self.mInputTrees

            for tree in trees:
                if self.mPruneTree:
                    taxa = self.mMapInput2Phylip.keys()
                    TreeTools.PruneTree( tree, taxa )

                taxa = TreeTools.GetTaxa( tree )
                self.updateMaps( taxa )
                TreeTools.MapTaxa( tree, self.mMapInput2Phylip )

                ## check if taxa are unique
                taxa = tree.get_taxa()
                staxa = set()
                
                skip = False
                for t in taxa:
                    if t in staxa:
                        if self.mLogLevel >= 1:
                            print "# skipping tree %s because of duplicate taxa." % (tree.name)
                        skip = True
                    staxa.add(t)
                    
                if skip:
                    continue
                
                outfile.write( TreeTools.Tree2Newick( tree ) + "\n" )
                self.mNInputTrees += 1
                
                if self.mLogLevel >= 1:
                    print "# written input tree with %i taxa to %s" % (len(TreeTools.GetTaxa( tree )), self.mTempdir + "/intree" )
                    print "#", TreeTools.Tree2Newick( tree ) 

            outfile.close()

        ## prepare input multiple alignment
        if self.mInputMali:

            if self.mInputMatrix:
                raise "both mali and matrix supplied - infile conflict."
            
            outfile = open(self.mTempdir + "/infile", "w" )

            identifiers = self.mInputMali.getIdentifiers()
            self.updateMaps( identifiers )
            self.mInputMali.mapIdentifiers( self.mMapInput2Phylip )
            self.mInputMali.writeToFile( outfile, format="phylip" )

            outfile.close()

            if self.mLogLevel >= 1:
                print "# written input multiple alignments with %i taxa and with %i to %s" %\
                      (self.mInputMali.getLength(), self.mInputMali.getWidth(), self.mTempdir + "/intree" )

    def run( self ):

        self.prepareRun()

        if not self.mProgram:
            raise UsageError( "no program specified." )

        s = subprocess.Popen( "%s" % (self.mProgram),
                              shell = True,
                              stdin = subprocess.PIPE,
                              stdout = subprocess.PIPE,
                              stderr = subprocess.PIPE,
                              cwd = self.mTempdir,
                              close_fds = True)                              

        (out, err) = s.communicate( "\n".join(self.mOptions) + "\n")

        if s.returncode != 0:
            raise UsageError, "Error in running phylip.\n%s\n%s\nTemporary directory was %s" % (out, err, self.mTempdir)

        ## Parse output files that might have been created:
        result = PhylipResult()

        ## parse tree file
        if os.path.exists( "%s/outtree" % self.mTempdir):
            
            nexus = TreeTools.Newick2Nexus( open("%s/outtree" % self.mTempdir, "r") )
            for tree in nexus.trees:
                TreeTools.MapTaxa( tree, self.mMapPhylip2Input )
            result.mNexus = nexus
            if self.mLogLevel >= 1:
                print "# received tree with %i taxa" % (len(TreeTools.GetTaxa(nexus.trees[0])))
                
        elif os.path.exists( "%s/outfile" % self.mTempdir ):
            
            if self.mProgram in ("dnadist", "protdist" ):
                infile = open( "%s/outfile" % self.mTempdir, "r")
                result.mMatrix, row_headers, col_headers = MatlabTools.readMatrix( infile, format="phylip" )
                result.mRowHeaders = []
                for x in row_headers: result.mRowHeaders.append( self.mMapPhylip2Input[ x ] )
                result.mColHeaders = result.mRowHeaders
            elif self.mProgram == "contrast":

                infile = open( "%s/outfile" % self.mTempdir, "r" )
                result.parseContrasts( infile )
                infile.close()

        else:
            raise "other return types not implemented"

        if self.mLogLevel >= 2:
            print out

        if self.mLogLevel == 0:
            shutil.rmtree( self.mTempdir )

        return result

if __name__ == "__main__":
    
    phylip = Phylip()

    parser = E.OptionParser( version = "%prog version: $Id: WrapperPhylip.py 2784 2009-09-10 11:41:14Z andreas $" )

    parser.add_option("-t", "--filename-input-tree", dest="filename_input_tree", type="string",
                      help="filename with tree information."  )
    parser.add_option("-T", "--filename-output-tree", dest="filename_output_tree", type="string",
                      help="output filename with tree information."  )
    parser.add_option("-p", "--program", dest="program", type="string",
                      help="program to use."  )
    parser.add_option("-o", "--options", dest="options", type="string",
                      help="input options."  )

    parser.set_defaults(
        filename_input_tree = None,
        filename_output_tree = None,
        program = None,
        options = "",
        )

    (options, args) = Experiment.Start( parser )

    if options.filename_input_tree != "-":
        phylip.setTree( open(options.filename_input_tree, "r").readlines() )
    elif options.filename_input_tree == "-":
        phylip.setTree( sys.stdin.readlines() )

    phylip.setOptions( options.options )
        
    phylip.setProgram( options.program )

    phylip.run()

    Experiment.Stop()
    



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
radar.py - 
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python radar.py --help

Type::

   python radar.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import optparse
import math
import time
import tempfile
import subprocess

USAGE="""python %s [OPTIONS] < mali > mali

reformat and edit a multiple alignment.

Methods are:

translate:      translate sequences.

mark-codons:    add space after each codon
pseudo-codons:  translate, but keep register with codons
interleaved-codons: mix amino acids and codons

Parameters are given to the option parameters in a comma-separated list in the order
that the edit operations are called upon.

""" % sys.argv[0]

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Genomics as Genomics
import CGAT.FastaIterator as FastaIterator

##------------------------------------------------------------
if __name__ == '__main__':

    parser = E.OptionParser( version = "%prog version: $Id: radar.py 2782 2009-09-10 11:40:29Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-m", "--method", dest="method", type="choice",
                      choices=("translate", "translate-till-stop",
                               "truncate-at-stop",
                               "mark-codons", "build-map",
                               "pseudo-codons", "interleaved-codons",
                               "remove-gaps",
                               "mask-seg", "mask-bias", "mask-codons",
                               "upper", "lower" ),
                      help="method to apply to sequences."  )
    
    parser.add_option("-p", "--parameters", dest="parameters", type="string",
                      help="parameter stack for methods that require one."  )

    parser.add_option("-t", "--type", dest="type", type="choice",
                      choices = ("aa", "na"),
                      help="sequence type (aa or na)."  )

    parser.set_defaults(
        only_codons = False,
        methods = "",
        parameters = "",
        type = "na",
        aa_mask_chars = "xX",
        aa_mask_char = "x",
        na_mask_chars = "nN",
        na_mask_char = "n",
        gap_chars = "-.",
        )

    (options, args) = E.Start( parser )
    
    iterator = FastaIterator.FastaIterator( sys.stdin)

    nseq = 0

    step_size = 3

    while 1:
        cur_record = iterator.next()
        
        if cur_record is None: break
        nseq += 1

        sequence = re.sub( " ", "", cur_record.sequence)
        l = len(sequence)

        if options.loglevel >= 1:
            options.stdlog.write("# analysing %s : length = %i\n" % (cur_record.title, l))

        nidentities = [0] * l

        for x in range(0, l, step_size):
            print x
            for d in range(0, l-x, step_size):
                for s1 in sequence[x:l-d]:
                    for s2 in sequence[x+d:l]:
                        if s1 == s2:
                            nidentities[d] += 1

        for x in range(l):
            print x, distances[x]
                
    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
csv_intersection.py - intersect two tables
======================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python csv_intersection.py --help

Type::

   python csv_intersection.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import getopt
import time
import optparse
import math
import tempfile

import CGAT.Experiment as E
import CGAT.CSV as CSV
import csv
import hashlib



class UniqueBuffer:
    mKeys = {}
    def __init__(self, outfile):
        self.mOutfile = outfile
    def write( self, out ):
        key = hashlib.md5(out).digest()
        if key not in self.mKeys:
            self.mKeys[key] = True
            self.mOutfile.write(out)
    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: csv_intersection.py 2782 2009-09-10 11:40:29Z andreas $")

    parser.add_option( "-u", "--unique", dest="unique", action="store_true",
                       help="output rows are uniq." )

    parser.set_defaults(
        remove = False,
        unique = False,
        )

    (options, args) = E.Start( parser, add_csv_options  = True)

    if len(args) != 2:
        raise "please specify two files to join."
        
    options.filename1, options.filename2 = args

    table1 = CSV.ReadTable( open(options.filename1, "r") )
    table2 = CSV.ReadTable( open(options.filename2, "r") )    

    if options.unique:
        outfile = UniqueBuffer(sys.stdout)
    else:
        outfile = options.stdout

    ## build new field list
    new_fields = []

    for x in options.join_fields1:
        new_fields.append( x )

    for x in fields1:
        if x not in options.join_fields1:
            new_fields.append( x )
        if x not in options.join_fields2:
            new_fields.append( x )
            
        writer = csv.DictWriter( outfile,
                                 fields,
                                 dialect=options.csv_dialect,
                                 lineterminator = options.csv_lineterminator,
                                 extrasaction = 'ignore' )
        

    
    if len(lines) > 0:
        
        old_fields = lines[0][:-1].split("\t")

        if options.remove:
            fields = []
            for x in old_fields:
                if x not in input_fields:
                    fields.append( x )
        else:
            fields = input_fields

        reader = csv.DictReader( lines,
                                 dialect=options.csv_dialect )
        
        print "\t".join(fields)        

        first_row = True
        for row in reader:
            row = CSV.ConvertDictionary( row )
            writer.writerow(row)

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
GO.py - compute GO enrichment from gene lists
=============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Usage
-----

The script ``GO.py`` will test for enrichment or depletion 
of GO categories within a gene list.

The script uses a hypergeometric test to check if a particular
GO category is enriched in a foreground set with respect
to a background set. Multiple testing is controlled by 
computing a empirical false discovery rate using a sampling 
procedure.

A GO analysis proceeds in three steps:

   1. building gene to GO assignments
   2. create one or more gene lists with foreground and background
   3. run one or more GO analyses for each of the foreground gene lists

This script analyses multiple gene lists in parallel when a matrix of
gene lists is provided. If multiple gene lists are provided, the FDR is 
controlled per gene list and not overall. However, my intuition is that
if the number of tests is large the results should be comparable as if 
the FDR was controlled globally, though I have no proof for this.

Building gene to GO assignments
+++++++++++++++++++++++++++++++

The easiest way to obtain a map from gene identifiers to GO assignments
is to down download GO assignments from the ENSEMBL database. The command
below will download go assignments for the human gene set
and save it in the file :file:`gene2go.data`::

   python GO.py 
      --filename-dump=gene2go.data
      --host=ensembldb.ensembl.org
      --user=anonymous 
      --database=homo_sapiens_core_54_36p
      --port=5306 
   > gene2go.log

In order to use GOslim categories, an additional mapping step needs to be performed.
The sequence of commands is::

    wget http://www.geneontology.org/GO_slims/goslim_goa.obo
    wget http://www.geneontology.org/ontology/gene_ontology.obo
    map2slim -outmap go2goslim.map goslim_goa.obo gene_ontology.obo
    python GO.py 
                --go2goslim 
                --filename-ontology=gene_ontology.obo 
                --slims=go2goslim.map 
                --log=goslim.log 
        < gene2go.data > gene2goslim.data

The first two commands obtain GOslim information. 
`map2slim <http://search.cpan.org/~cmungall/go-perl/scripts/map2slim>`_
is part of Chris Mungall's `go-perl <http://search.cpan.org/~cmungall/go-perl/>`_ module 
and the last command converts the gene-to-GO assignment into gene-to-GOSlim assignments.

The gene-to-GO mapping can be constructed any other way. It is simply 
a table of tab-separated values::

   go_type gene_id go_id   description     evidence
   biol_process    ENSG00000151729 GO:0000002      mitochondrial genome maintenance        NA
   biol_process    ENSG00000025708 GO:0000002      mitochondrial genome maintenance        NA
   biol_process    ENSG00000115204 GO:0000002      mitochondrial genome maintenance        NA
   ...

Building gene lists
+++++++++++++++++++

GO requires a list of genes to test for enrichment. This list is simply
a table with one column of gene identifiers. For example::

   gene_id
   ENSG00000116586
   ENSG00000065809
   ENSG00000164048
   ENSG00000115137
   ENSG00000121210

Alternatively, the gene list can be a multi-column table such as::

   gene_id             dataset1    dataset2
   ENSG00000116586     1           0
   ENSG00000065809     0           0
   ENSG00000164048     1           0
   ENSG00000115137     1           1
   ENSG00000121210     0           1

In this case, enrichment is computed for multiple datasets at once. Make sure
to add the ``%(set)s`` place holder to ``--filename-output-pattern``.

If no background is given, all genes that have GO assignments will constitute
the background. 

Statistics
++++++++++

Enrichment is computed using the hypergeometric test. 

.. todo::
    * apply filtering
    * more stats
    * more FDR

Running the GO analysis
+++++++++++++++++++++++

The command below runs a GO analysis, computing an FDR using 10.000 samples::

    python GO.py 
        --filename-input=gene2go.data
        --genes=foreground
        --background=background 
        --sample=10000
        --fdr 
        --filename-ontology=gene_ontology.obo
        --output-filename-pattern='result/%(set)s.%(go)s.%(section)s' 
   > go.log

The output will be stored in the directory :file:`result` and output files will be
created according to the pattern ``<set>.<go>.<section>``. ``<set>`` is the gene set
that is analysed, ``<go>`` is one of ``biol_process``, ``mol_function`` and ``cell_location``.
``<section>`` denotes the file contents. Files output are:

+------------+----------------------------------------------+
|``section`` | contents                                     |
+------------+----------------------------------------------+
|samples     |sampling statistics                           |
+------------+----------------------------------------------+
|overall     |table with full results                       |
+------------+----------------------------------------------+
|results     |table with only the significant results       |
+------------+----------------------------------------------+
|parameters  |input and sampling parameters                 |
+------------+----------------------------------------------+
|fg          |assigments for genes in the foreground set    |
+------------+----------------------------------------------+

Other options
+++++++++++++

The script can accept other ontologies than just GO ontologies. 

Code
----

'''
import os, sys, string, re, getopt, time, optparse, math, tempfile, subprocess, random
import collections

import scipy
import scipy.stats
import scipy.special
import numpy
import CGAT.Stats as Stats
import CGAT.Database as Database
import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.CSV as CSV
from CGAT.GO import *

##---------------------------------------------------------------------------    
def main():

    parser = E.OptionParser( version = "%prog version: $Id: GO.py 2883 2010-04-07 08:46:22Z andreas $", usage = globals()["__doc__"])

    dbhandle = Database.Database()
    
    parser.add_option("-s", "--species", dest="species", type="string",
                      help="species to use [default=%default]." )

    parser.add_option("-i", "--slims", dest="filename_slims", type="string",
                      help="filename with GO SLIM categories [default=%default].")

    parser.add_option( "-g", "--genes", dest="filename_genes", type="string",
                       help="filename with genes to analyse [default=%default]." )

    parser.add_option( "-b", "--background", dest="filename_background", type="string",
                       help="filename with background genes to analyse [default=%default]." )

    parser.add_option( "-m", "--minimum-counts", dest="minimum_counts", type="int",
                       help="minimum count - ignore all categories that have fewer than # number of genes"
                            " [default=%default]." )

    parser.add_option( "-o", "--sort-order", dest="sort_order", type="choice",
                       choices=("fdr", "pvalue", "ratio" ),
                       help="output sort order [default=%default]." )

    parser.add_option( "--ontology", dest="ontology", type="string", action="append",
                       help="go ontologies to analyze. Ontologies are tested separately."
                       " [default=%default]." )

    parser.add_option( "-t", "--threshold", dest="threshold", type="float",
                       help="significance threshold [>1.0 = all ]. If --fdr is set, this refers to the fdr, otherwise it is a cutoff for p-values." )

    parser.add_option ("--filename-dump", dest="filename_dump", type="string",
                       help="dump GO category assignments into a flatfile [default=%default]." )

    parser.add_option ("--filename-gene2name", dest="filename_gene2name", type="string",
                       help="optional filename mapping gene identifiers to gene names [default=%default]." )

    parser.add_option ("--filename-ontology", dest="filename_ontology", type="string",
                       help="filename with ontology in OBO format [default=%default]." )

    parser.add_option ( "--filename-input", dest="filename_input", type="string",
                       help="read GO category assignments from a flatfile [default=%default]." )

    parser.add_option ( "--sample", dest="sample", type="int",
                        help="do sampling (with # samples) [default=%default]." )

    parser.add_option ( "--filename-output-pattern", "--output-filename-pattern", 
                        dest = "output_filename_pattern", type="string",
                        help="pattern with output filename pattern (should contain: %(go)s and %(section)s ) [default=%default]")

    parser.add_option ( "--fdr", dest="fdr", action="store_true",
                        help="calculate and filter by FDR [ReadGene2GOFromFiledefault=%default]." )
    
    parser.add_option ( "--go2goslim", dest="go2goslim", action="store_true",
                        help="convert go assignments in STDIN to goslim assignments and write to STDOUT [default=%default]." )

    parser.add_option ( "--gene-pattern", dest = "gene_pattern", type="string",
                        help="pattern to transform identifiers to GO gene names [default=%default].")
    
    parser.add_option( "--filename-map-slims", dest="filename_map_slims", type="string",
                       help="write mapping between GO categories and GOSlims [default=%default].")

    parser.add_option( "--get-genes", dest="get_genes", type="string",
                       help="list all genes in the with a certain GOID [default=%default]." )

    parser.add_option( "--strict", dest="strict", action="store_true",
                       help="require all genes in foreground to be part of background. "
                       "If not set, genes in foreground will be added to the background [default=%default]." )

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices = ( "empirical", "storey", "BH" ),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default]."  )


    parser.add_option( "--pairwise", dest="compute_pairwise", action="store_true",
                       help="compute pairwise enrichment for multiple gene lists. "
                       "[default=%default]." )

    # parser.add_option( "--qvalue-lambda", dest="qvalue_lambda", type="float",
    #                   help="fdr computation: lambda [default=%default]."  )

    # parser.add_option( "--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
    #                    choices = ("smoother", "bootstrap" ),
    #                    help="fdr computation: method for estimating pi0 [default=%default]."  )
    
    parser.set_defaults( species = None,
                         filename_genes = "-",
                         filename_background = None,
                         filename_slims = None,
                         minimum_counts = 0,
                         ontology = [],
                         filename_dump = None,
                         sample = 0,
                         fdr = False,
                         output_filename_pattern = None,
                         threshold = 0.05,
                         filename_map_slims = None,
                         gene_pattern = None,
                         sort_order = "ratio",
                         get_genes = None,
                         strict = False,
                         qvalue_method = "empirical",
                         pairs_min_observed_counts = 3,
                         compute_pairwise = False,
                         filename_gene2name = None
                         )

    (options, args) = E.Start( parser, add_mysql_options = True )

    if options.go2goslim:
        convertGo2Goslim( options )
        E.Stop()
        sys.exit(0)

    if options.fdr and options.sample == 0:
        E.warn( "fdr will be computed without sampling" )
        
    #############################################################
    ## dump GO
    if options.filename_dump:
        # set default orthologies to GO
        if not options.ontology:
            options.ontology = ["biol_process", "mol_function", "cell_location"] 

        E.info( "dumping GO categories to %s" % (options.filename_dump) )

        dbhandle.Connect( options )
            
        outfile = IOTools.openFile( options.filename_dump, "w", create_dir = True )
        DumpGOFromDatabase( outfile,
                            dbhandle,
                            options )
        outfile.close()
        E.Stop()
        sys.exit(0)

    #############################################################
    ## read GO categories from file
    if options.filename_input:
        E.info( "reading association of categories and genes from %s" % (options.filename_input) )
        infile = IOTools.openFile(options.filename_input)
        gene2gos, go2infos = ReadGene2GOFromFile( infile )
        infile.close()

    if options.filename_gene2name:
        E.info("reading gene identifier to gene name mapping from %s" % options.filename_gene2name)
        infile = IOTools.openFile( options.filename_gene2name)
        gene2name = IOTools.readMap( infile, has_header = True )
        infile.close()
        E.info("read %i gene names for %i gene identifiers" % (len(set(gene2name.values())),
                                                               len(gene2name)))
    else:
        gene2name = None

    #############################################################
    ## read GO ontology from file
    if options.filename_ontology:
        E.info( "reading ontology from %s" % (options.filename_ontology) )
        
        infile = IOTools.openFile(options.filename_ontology)
        ontology = readOntology( infile )
        infile.close()
        
        def _g():
            return collections.defaultdict( GOInfo )
        go2infos = collections.defaultdict( _g );

        ## substitute go2infos
        for go in ontology.values():
            go2infos[go.mNameSpace][go.mId] = GOInfo( go.mId,
                                                      go_type = go.mNameSpace,
                                                      description = go.mName )

    #############################################################
    ## get foreground gene list
    input_foreground, genelists = ReadGeneLists( options.filename_genes, 
                                                 gene_pattern = options.gene_pattern )

    E.info( "read %i genes for forground in %i gene lists" % (len(input_foreground), len(genelists)) )

    #############################################################
    ## get background
    if options.filename_background:
        
        # nick - bug fix: background is the first tuple element from ReadGeneLists
        input_background = ReadGeneLists( options.filename_background, 
                                          gene_pattern = options.gene_pattern )[0]
        E.info( "read %i genes for background" % len(input_background) )
    else:
        input_background = None

    #############################################################
    ## sort out which ontologies to test
    if not options.ontology: 
        if options.filename_input:
            options.ontology = gene2gos.keys()

    E.info( "found %i ontologies: %s" % (len(options.ontology), options.ontology))

    outfile_summary = options.stdout
    outfile_summary.write( "\t".join( (
                "genelist",
                "ontology",
                "significant",
                "threshold",
                "ngenes",
                "ncategories",
                "nmaps",
                "nforegound",
                "nforeground_mapped",
                "nbackground",
                "nbackground_mapped",
                "nsample_counts",
                "nbackground_counts",
                "psample_assignments",
                "pbackground_assignments") ) + "\n" )

    #############################################################
    ## get go categories for genes
    for test_ontology in options.ontology:

        # store results for aggregate output of multiple gene lists
        all_results = []
        all_significant_results = []
        all_genelists_with_results = []

        E.info( "working on ontology %s" % test_ontology )
        #############################################################
        ## get/read association of GO categories to genes
        if options.filename_input:
            gene2go, go2info = gene2gos[test_ontology], go2infos[test_ontology]
        else:
            E.info( "reading data from database ..." )

            dbhandle.Connect( options )
            gene2go, go2info = ReadGene2GOFromDatabase( dbhandle,
                                                        test_ontology,
                                                        options.database, options.species )

            E.info( "finished" )

        if len(go2info) == 0:
            E.warn( "could not find information for terms - could be mismatch between ontologies")

        ngenes, ncategories, nmaps, counts_per_category = CountGO( gene2go )        
        E.info( "assignments found: %i genes mapped to %i categories (%i maps)" % (ngenes, ncategories, nmaps) )

        if options.minimum_counts > 0:
            to_remove = set([ x for x,y in counts_per_category.iteritems() if y < options.minimum_counts ])
            E.info("removing %i categories with less than %i genes" % (len(to_remove), options.minimum_counts ) )
            removeCategories( gene2go, to_remove )

            ngenes, ncategories, nmaps, counts_per_category = CountGO( gene2go )        
            E.info( "assignments after filtering: %i genes mapped to %i categories (%i maps)" % (ngenes, ncategories, nmaps) )

        for genelist_name, foreground in genelists.iteritems():

            msgs = []
            E.info("processing %s with %i genes" % (genelist_name, len(foreground)))
            ##################################################################
            ##################################################################
            ##################################################################
            ## build background - reconcile with foreground
            ##################################################################
            if input_background == None:
                background = list(gene2go.keys())
            else:
                background = list(input_background)

            # nick - bug-fix backgorund included the foreground in a tuple. 
            # background is the first tuple element 
            missing = foreground.difference(set(background))

            if options.strict:
                assert len(missing) == 0, \
                    "%i genes in foreground but not in background: %s" % (len(missing), str(missing))
            else:
                if len(missing) != 0:
                    E.warn( "%i genes in foreground that are not in background - added to background of %i" %\
                                (len(missing), len(background)) )

                background.extend(missing)

            E.info( "(unfiltered) foreground=%i, background=%i" % (len(foreground), len(background)))

            #############################################################
            ## sanity checks:            
            ## are all of the foreground genes in the dataset
            ## missing = set(genes).difference( set(gene2go.keys()) )
            ## assert len(missing) == 0, "%i genes in foreground set without GO annotation: %s" % (len(missing), str(missing))

            #############################################################            
            ## read GO slims and map GO categories to GO slim categories
            if options.filename_slims:
                go_slims = GetGOSlims( IOTools.openFile(options.filename_slims, "r") )

                if options.loglevel >=1:
                    v = set()
                    for x in go_slims.values():
                        for xx in x: v.add(xx)
                    options.stdlog.write( "# read go slims from %s: go=%i, slim=%i\n" %\
                                              ( options.filename_slims,
                                                len(go_slims), 
                                                len( v ) ))



                if options.filename_map_slims:
                    if options.filename_map_slims == "-":
                        outfile = options.stdout
                    else:
                        outfile=IOTools.openFile(options.filename_map_slims, "w" )

                    outfile.write( "GO\tGOSlim\n" )
                    for go, go_slim in go_slims.items():
                        outfile.write("%s\t%s\n" % (go, go_slim))

                    if outfile != options.stdout:
                        outfile.close()

                gene2go = MapGO2Slims( gene2go, go_slims, ontology = ontology )

                if options.loglevel >=1:
                    ngenes, ncategories, nmaps, counts_per_category = CountGO( gene2go )
                    options.stdlog.write( "# after go slim filtering: %i genes mapped to %i categories (%i maps)\n" % (ngenes, ncategories, nmaps) )

            #############################################################
            ## Just dump out the gene list
            if options.get_genes:
                fg, bg, ng = [], [], []

                for gene, vv in gene2go.items():
                    for v in vv:
                        if v.mGOId == options.get_genes:
                            if gene in genes:
                                fg.append( gene )
                            elif gene in background:
                                bg.append( gene )
                            else:
                                ng.append( gene )

                ## skip to next GO class
                if not (bg or ng): continue

                options.stdout.write( "# genes in GO category %s\n" % options.get_genes )
                options.stdout.write( "gene\tset\n" )
                for x in fg: options.stdout.write("%s\t%s\n" % ("fg", x))
                for x in bg: options.stdout.write("%s\t%s\n" % ("bg", x))           
                for x in ng: options.stdout.write("%s\t%s\n" % ("ng", x))                       

                E.info( "nfg=%i, nbg=%i, nng=%i" % (len(fg), len(bg), len(ng) ))

                E.Stop()
                sys.exit(0)

            #############################################################
            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'foreground',
                                   set = genelist_name )

            outfile.write ("gene_id\n%s\n" % ("\n".join( sorted( foreground) ) ) )
            if options.output_filename_pattern:
                outfile.close()

            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'background',
                                   set = genelist_name )

            outfile.write ("gene_id\n%s\n" % ("\n".join( sorted( background[0]) ) ) )
            if options.output_filename_pattern:
                outfile.close()

            #############################################################
            ## do the analysis
            go_results = AnalyseGO( gene2go, foreground, background )

            if len(go_results.mSampleGenes) == 0:
                E.warn( "%s: no genes with GO categories - analysis aborted" % genelist_name)
                continue

            pairs = go_results.mResults.items()

            #############################################################
            ## calculate fdr for each hypothesis
            if options.fdr:
                fdrs, samples, method  = computeFDRs( go_results, options, test_ontology )
                for x,v in enumerate(pairs):
                    v[1].mQValue = fdrs[v[0]][0]
            else:
                fdrs, samples, method = {}, None
                
            msgs.append( "fdr=%s" % method)

            if options.sort_order == "fdr":
                pairs.sort( lambda x, y: cmp(x[1].mQValue, y[1].mQValue))           
            elif options.sort_order == "ratio":
                pairs.sort( lambda x, y: cmp(x[1].mRatio, y[1].mRatio))
            elif options.sort_order == "pvalue":
                pairs.sort( lambda x, y: cmp(x[1].mPValue, y[1].mPValue))

            #############################################################
            #############################################################
            #############################################################
            ## output the full result
            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'overall',
                                   set = genelist_name )

            outputResults( outfile, pairs, go2info, options, fdrs = fdrs, samples = samples )

            if options.output_filename_pattern:
                outfile.close()

            #############################################################
            #############################################################
            #############################################################
            # filter significant results and output
            filtered_pairs = selectSignificantResults( pairs, fdrs, options )

            nselected = len(filtered_pairs)
            nselected_up = len( [x for x in filtered_pairs if x[1].mRatio > 1 ] )
            nselected_down = len( [x for x in filtered_pairs if x[1].mRatio < 1 ] )
            
            assert nselected_up + nselected_down == nselected

            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'results',
                                   set = genelist_name )

            outputResults( outfile, 
                           filtered_pairs, 
                           go2info, 
                           options,
                           fdrs = fdrs, 
                           samples = samples )
            
            if options.output_filename_pattern:
                outfile.close()

            #############################################################
            #############################################################
            #############################################################
            # save results for multi-gene-list analysis
            all_results.append( pairs )
            all_significant_results.append( filtered_pairs )
            all_genelists_with_results.append( genelist_name )

            #############################################################
            #############################################################
            #############################################################
            ## output parameters
            ngenes, ncategories, nmaps, counts_per_category = CountGO( gene2go )

            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'parameters',
                                   set = genelist_name )

            nbackground = len(background)
            if nbackground == 0:
                nbackground = len(go_results.mBackgroundGenes)

            outfile.write( "# input go mappings for gene list '%s' and category '%s'\n" % (genelist_name, test_ontology ))
            outfile.write( "parameter\tvalue\tdescription\n" )
            outfile.write( "mapped_genes\t%i\tmapped genes\n" % ngenes )
            outfile.write( "mapped_categories\t%i\tmapped categories\n" % ncategories )
            outfile.write( "mappings\t%i\tmappings\n" % nmaps )
            outfile.write( "genes_in_fg\t%i\tgenes in foreground\n" % len(foreground) )
            outfile.write( "genes_in_fg_with_assignment\t%i\tgenes in foreground with GO assignments\n" % (len(go_results.mSampleGenes)) )
            outfile.write( "genes_in_bg\t%i\tinput background\n" % nbackground )
            outfile.write( "genes_in_bg_with_assignment\t%i\tgenes in background with GO assignments\n" % (len(go_results.mBackgroundGenes)) )
            outfile.write( "associations_in_fg\t%i\tassociations in sample\n"     % go_results.mSampleCountsTotal )
            outfile.write( "associations_in_bg\t%i\tassociations in background\n" % go_results.mBackgroundCountsTotal )
            outfile.write( "percent_genes_in_fg_with_association\t%s\tpercent genes in sample with GO assignments\n" % (IOTools.prettyPercent( len(go_results.mSampleGenes) , len(foreground), "%5.2f" )))
            outfile.write( "percent_genes_in_bg_with_associations\t%s\tpercent genes background with GO assignments\n" % (IOTools.prettyPercent( len(go_results.mBackgroundGenes), nbackground, "%5.2f" )))
            outfile.write( "significant\t%i\tsignificant results reported\n" % nselected )
            outfile.write( "significant_up\t%i\tsignificant up-regulated results reported\n" % nselected_up )
            outfile.write( "significant_down\t%i\tsignificant up-regulated results reported\n" % nselected_down )
            outfile.write( "threshold\t%6.4f\tsignificance threshold\n" % options.threshold )        

            if options.output_filename_pattern:
                outfile.close()

            if outfile_summary:
                outfile_summary.write( "\t".join( map(str, ( \
                                genelist_name,
                                test_ontology,
                                nselected,
                                options.threshold,
                                ngenes,
                                ncategories,
                                nmaps,
                                len(foreground),
                                len(go_results.mSampleGenes),
                                nbackground,
                                len(go_results.mBackgroundGenes),
                                go_results.mSampleCountsTotal,
                                go_results.mBackgroundCountsTotal,
                                IOTools.prettyPercent( len(go_results.mSampleGenes) , len(foreground), "%5.2f" ),
                                IOTools.prettyPercent( len(go_results.mBackgroundGenes), nbackground, "%5.2f" ),
                                ",".join( msgs) ) ) ) + "\n" )

            #############################################################
            #############################################################
            #############################################################
            ## output the fg patterns
            outfile = getFileName( options, 
                                   go = test_ontology,
                                   section = 'withgenes',
                                   set = genelist_name )

            outputResults( outfile, pairs, go2info, options, 
                           fdrs = fdrs, 
                           gene2go = gene2go,
                           foreground = foreground,
                           gene2name = gene2name )

            if options.output_filename_pattern:
                outfile.close()

        if len(genelists) > 1:

            ######################################################################
            ######################################################################
            ######################################################################
            ## output various summary files
            ## significant results
            outputMultipleGeneListResults( all_significant_results, 
                                           all_genelists_with_results, 
                                           test_ontology, 
                                           go2info,
                                           options,
                                           section = 'significant')

            ## all results
            outputMultipleGeneListResults( all_results, 
                                           all_genelists_with_results, 
                                           test_ontology, 
                                           go2info,
                                           options,
                                           section = 'all')

            
            if options.compute_pairwise:
                pairwiseGOEnrichment( all_results,
                                      all_genelists_with_results,
                                      test_ontology,
                                      go2info,
                                      options )

    E.Stop()


if __name__ == "__main__":
    sys.exit( main() )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
psl2table.py - output stats for psl formatted alignments
================================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

.. todo::
   
   describe purpose of the script.

Usage
-----

Example::

   python psl2table.py --help

Type::

   python psl2table.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import math

USAGE = \
"""analyze sequence pairs from a psl formatted table.

The sequences are assumed to be nucleotide sequences.

Methods available are:

baseml: compute baseml rates 
counts: compute residue counts (percent G+C, ...)
match:  compute match statistics (pid, coverage)
"""

import CGAT.Experiment as E
import CGAT.Blat as Blat
import CGAT.SequenceProperties as SequenceProperties
import CGAT.SequencePairProperties as SequencePairProperties
import CGAT.IOTools as IOTools
import CGAT.WrapperCodeML as WrapperCodeML

##---------------------------------------------------------------------------------------------
class Counter:
    def __init__( self, options):
        pass

class CounterMatch( Counter ):
    
    def __init__(self,*args,**kwargs):
        Counter.__init__(self,*args,**kwargs)

    def __call__(self, match):
        
        self.mPid = 100.0 * match.mNMatches / (match.mNMatches + match.mNMismatches )
        self.mQueryCoverage = 100.0 * (match.mNMatches + match.mNMismatches) / match.mQueryLength

    def __str__( self ):
        return "%6.4f\t%6.4f" % (self.mPid, self.mQueryCoverage)

    def getHeaders(self):
        return [ "pid", "qCov" ]

class QueriesCounter( SequenceProperties.SequencePropertiesNA ):
    def __init__(self, *args, **kwargs):
        SequenceProperties.SequencePropertiesNA.__init__(self, *args, **kwargs)

    def __call__( self, seq1, seq2 ):
        SequenceProperties.SequencePropertiesNA.loadSequence( self, seq1 )

class SbjctsCounter( SequenceProperties.SequencePropertiesNA ):
    def __init__(self, *args, **kwargs):
        SequenceProperties.SequencePropertiesNA.__init__(self, *args, **kwargs)

    def __call__( self, seq1, seq2 ):
        SequenceProperties.SequencePropertiesNA.loadSequence( self, seq2 )



##---------------------------------------------------------------------------------------------
    
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: psl2table.py 2891 2010-04-07 08:59:18Z andreas $",
                                    usage = globals()["__doc__"])

    parser.add_option( "--mask-lowercase", dest="mask_lowercase", action="store_true",
                      help="mask lowercase characters before computing properties [default=%default]" )

    parser.add_option( "--with-match", dest="with_match", action="store_true",
                      help="echo the match in output [default=%default]" )

    parser.add_option( "--without-match", dest="with_match", action="store_false",
                      help="do not echo the match in output [default=%default]" )

    parser.add_option("-m", "--method", dest="methods", type="choice", action="append",
                      choices=("counts", "baseml", "match", "query-counts", "sbjct-counts" ),
                      help="methods to compute properties between sequence pairs." )

    WrapperCodeML.BaseML().AddOptions( parser )

    parser.set_defaults(
        methods = [],
        mask_lowercase = False,
        is_pslx = True,
        with_match = True,
        )
    
    (options, args) = E.Start( parser )

    counters_plain = []
    counters = []

    for method in options.methods:
        if method == "counts":
            counters.append( SequencePairProperties.SequencePairPropertiesCountsNa() )
        elif method == "query-counts":
            counters.append( QueriesCounter() )
        elif method == "sbjct-counts":
            counters.append( SbjctsCounter() )
        elif method == "baseml":
            counters.append( SequencePairProperties.SequencePairPropertiesBaseML( options ) )
        elif method == "match":
            counters_plain.append( CounterMatch( options ) )
            
    if counters:
        iterator = Blat.iterator_pslx( options.stdin )
        header = "\t".join(Blat.MatchPSLX().getHeaders())
    else:
        iterator = Blat.iterator( options.stdin )
        header = "\t".join(Blat.Match().getHeaders())

    if not options.with_match:
        header = "qName"

    options.stdout.write( "\t".join( 
            [header,] + 
            [ "\t".join(x.getHeaders()) for x in counters] +
            [ "\t".join(x.getHeaders()) for x in counters_plain] ) + "\n" )

    ninput, noutput, nskipped = 0, 0, 0

#     ## setup totals
#     totals = {}
#     for section in options.sections:
#         if section == "length":
#             s = SequencePropertiesLength()
#         elif section == "na":
#             s = SequencePropertiesNA()
#         elif section == "aa":
#             s = SequencePropertiesAA()
#         elif section == "degeneracy":
#             s = SequencePropertiesDegeneracy()
#         elif section == "bias":
#             s = SequencePropertiesBias( reference_codons )
#         elif section == "codons":
#             s = SequencePropertiesCodons()
#         elif section == "codon-usage":
#             s = SequencePropertiesCodonUsage()
#         elif section == "codon-translator":
#             s = SequencePropertiesCodonTranslator()
#         else:
#             raise "unknown section %s" % section
        
#         totals[section] = s


    for match in iterator:
        ninput += 1

        if options.with_match:
            options.stdout.write( str(match) )
        else:
            options.stdout.write( match.mQueryId )
            
        if counters:
        
            qseq = match.mQuerySequence 
            sseq = match.mSbjctSequence 

            # mask non printable characters - sometimes
            # appear after using pslToPslX
            qseq = [ re.sub( "[^a-zA-Z]", "N", x ) for x in qseq ]
            sseq = [ re.sub( "[^a-zA-Z]", "N", x ) for x in sseq ]

            if options.mask_lowercase:
                qseq = [ re.sub( "[a-z]", "N", x) for x in qseq ]
                sseq = [ re.sub( "[a-z]", "N", x) for x in sseq ]
                
            match.mQuerySequence = qseq
            match.mSbjctSequence = sseq

            qseq = "".join( match.mQuerySequence ).upper()
            sseq = "".join( match.mSbjctSequence ).upper()

            if len(qseq) != len(sseq):
                if options.loglevel >= 1:
                    options.stdlog.write( "# WARNING: two sequences of unequal length in match\n# %s\n" % str(match) )
                nskipped += 1
                continue

            for counter in counters: counter( qseq, sseq )

            options.stdout.write( "\t" + 
                                  "\t".join( \
                    [str( counter ) for counter in counters] ) )
        
        if counters_plain:

            for counter in counters_plain: counter( match )

            options.stdout.write( "\t" + 
                                  "\t".join( \
                    [str( counter ) for counter in counters_plain] ) )

        options.stdout.write( "\n" )

        noutput += 1

    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, nskipped=%i\n" % (ninput, noutput, nskipped) )

    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
cluster_distribute.py - distribute files to cluster nodes
=========================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script distributes one or more files to the local
scratch directory on all nodes in the cluster.

The script uses ``rsync`` to only copy files that are newer.

Usage
-----

Example::

   python cluster_distribute.py -collection=blast /net/cpp-group/tools/polyphen-2.0.18/nrdb/uniref100.*.{pin,psd,psi,phr,psq}

The above command mirrors uniprot blast indexed databases
into the directory :file:`/scratch/blast` on the nodes.

Type::

   python cluster_distribute.py --help

for command line help.

.. note::

   The directory needs to be cleaned up as disk space on the
   nodes is limited.

.. todo::
   
   Currently files are copied to all nodes. This is potentially
   wasteful if jobs will only be executed on a few nodes.

Documentation
-------------

Code
----

'''

import os
import sys
import re
import optparse

import CGAT.Experiment as E

def getNodes( nodes = None):
    '''hack - allow ranges, ...'''
    if nodes == None or len(nodes) == 0:
        # return [ "node%02i" % x for x in range(1,15) + range(20,34) + range(48,62) ]
        return [ "cgat%03i" % x for x in range( 1,10) + range(101,117) ] + \
            [ "cgat150", "cgatsmp1", "andromeda", "gandalf", "saruman" ]
    return nodes

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: cgat_script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-s", "--scratchdir", dest="scratchdir", type="string",
                      help="the scratch directory on the nodes [default=%default]."  )

    parser.add_option("-c", "--collection", dest="collection", type="string",
                      help="files will be put into collection. This is a directory that will be"
                      " created just below the scratch directory [default=%default]."  )

    parser.set_defaults(
        scratchdir = "/scratch",
        collection = "",
        nodes = [],
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv )

    if len(args) == 0:
        raise ValueError("please specify a collection of files/directories that should be mirrored." )

    targetdir = os.path.join( options.scratchdir, options.collection )

    nodes = getNodes( options.nodes )

    E.info( "copying to %s on nodes %s" % (targetdir, ",".join(nodes) ))

    ninput, noutput, nskipped = 0, 0, 0

    filenames = " ".join(args)

    for node in nodes:
        E.info( "copying to node %s" % node )
        ninput += 1
        statement = '''
               ssh %(node)s mkdir %(targetdir)s >& /dev/null;
               rsync --progress -az %(filenames)s %(node)s:%(targetdir)s
        ''' % locals()
        E.run( statement )
        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput,nskipped) )


    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



################################################################################
#   Gene prediction pipeline 
#
#   $Id: fasta2gff.py 2861 2010-02-23 17:36:32Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
fasta2gff.py - create random segments from fasta file
=====================================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script creates a random sample of "exons" from 
a fasta file.

Usage
-----

Example::

   python <script_name>.py --help

Type::

   python <script_name>.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os
import sys
import string
import re
import getopt
import tempfile
import time
import optparse
import math
import glob

import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.GTF as GTF

USAGE="""python %s [OPTIONS] 



Version: $Id: fasta2gff.py 2861 2010-02-23 17:36:32Z andreas $
""" % sys.argv[0]

def writeHeader( outfile ):
    outfile.write( "\t".join( ("contig",
                               "nresidues",
                               "ngaps",
                               "nseqregions",                                      
                               "ngapregions", 
                               "nA", "nC", "nG", "nT",
                               "nN", "nX", "nO" ) ) + "\n" )
    
    

def main():

    parser = E.OptionParser( version = "%prog version: $Id: fasta2gff.py 2861 2010-02-23 17:36:32Z andreas $")

    parser.add_option( "-g", "--genome-file", dest="genome_file", type="string",
                       help="filename with genome."  )

    parser.add_option("-a", "--as-gtf", dest="as_gtf", action="store_true",
                      help="output as gtf."  )

    parser.add_option( "-f", "--fragment-size", dest="fragment_size", type="int",
                       help="fixed size of fragments [default=%default]."  )

    parser.add_option( "-s", "--sample-size", dest="sample_size", type="int",
                       help="fixed size of fragments."  )

    parser.set_defaults(
        as_gtf = False,
        genome_file = None,
        fragment_size = 1000,
        sample_size = 10000,
        pattern_id = "%08i",
        )

    (options, args) = E.Start( parser )

    fasta = IndexedFasta.IndexedFasta( options.genome_file )
    contigs = fasta.getContigSizes()

    if options.as_gtf:
        entry = GTF.Entry()
    else:
        entry = GTF.Entry()

    n = 0
    entry.feature = "exon"
    entry.source = "random"

    for x in range( options.sample_size):

        entry.contig, entry.strand, entry.start, entry.end = fasta.getRandomCoordinates( options.fragment_size )
        if entry.strand == "-":
            l = contigs[entry.contig]
            entry.start, entry.end = l-entry.end, l-entry.start

        if options.as_gtf:
            entry.gene_id = options.pattern_id % n
            entry.transcript_id = entry.gene_id

        options.stdout.write( str(entry) + "\n" )
        n += 1

    E.Stop()

if __name__ == "__main__":
    sys.exit(main())



####
####
##
##
## Copyright (C) 2008 Andreas Heger All rights reserved
##
## Author: Andreas Heger <heger@ebi.ac.uk>
##
## $Id: bed2gff.py 2861 2010-02-23 17:36:32Z andreas $
##
##
####
####
"""
bed2gff.py - convert bed to gff/gtf
===================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script converts a :term:`bed`-formatted file to a :term:`gff` or :term:`gtf`-formatted file.

It aims to populate the appropriate fields in the :term:`gff` file with columns in the :term:`bed` file.

If ``--is-gtf`` is set and a name column in the :term:`bed` file is present, its contents will be set 
as ``gene_id`` and ``transcript_id``. Otherwise, a numeric ``gene_id`` or ``transcript_id`` will be set 
according to ``--id-format``.

Usage
-----

Example::

   python bed2gff.py < in.bed > out.gff

Type::

   python bed2gff.py --help

for command line help.

Command line options
--------------------

""" 
import sys
import re
import string
import optparse
import time
import os
import itertools
import tempfile
import subprocess
import shutil

import CGAT.Experiment as E
import CGAT.Stats as Stats
import CGAT.GTF as GTF
import CGAT.Bed as Bed
import CGAT.IndexedFasta as IndexedFasta
import CGAT.IOTools as IOTools

def main( argv = sys.argv ):

    parser = E.OptionParser( version = "%prog version: $Id: bed2gff.py 2861 2010-02-23 17:36:32Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-a", "--as-gtf", dest="as_gtf", action="store_true",
                      help="output as gtf."  )

    parser.add_option("-f", "--id-format", dest="id_format", type="string",
                      help="format for numeric identifier if --as-gtf is set and no name in bed file [%default]."  )

    parser.set_defaults( as_gtf = False,
                         id_format = "%08i",
                         test = None )
    
    (options, args) = E.Start( parser, add_pipe_options = True )

    as_gtf = options.as_gtf
    id_format = options.id_format

    if as_gtf:
        gff = GTF.Entry()
    else:
        gff = GTF.Entry()

    gff.source = "bed"
    gff.feature = "exon"

    ninput, noutput, nskipped = 0, 0, 0

    id = 0
    for bed in Bed.iterator( options.stdin ):

        ninput += 1

        gff.contig = bed.contig
        gff.start = bed.start 
        gff.end = bed.end
        if bed.fields and len(bed.fields) >= 3:
            gff.strand = bed.fields[2]
        else: 
            gff.strand = "."

        if bed.fields and len(bed.fields) >= 2:
            gff.score = bed.fields[1]
        
        if as_gtf:
            if bed.fields:
                gff.gene_id = bed.fields[0]
                gff.transcript_id = bed.fields[0]
            else:
                id += 1
                gff.gene_id = id_format % id
                gff.transcript_id = id_format % id            
        else:
            if bed.fields:
                gff.source = bed.fields[0]
            
        options.stdout.write( str(gff) + "\n" )

        noutput += 1

    E.info( "ninput=%i, noutput=%i, nskipped=%i" % (ninput, noutput, nskipped) )

    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv ) )



# cython: profile=True
#cimport csamtools
from pysam.csamtools cimport *

import collections, array, struct
import CGAT.Experiment as E

import os
import sys
import string
import re
import optparse
import math
import time
import tempfile
import subprocess
import types
import bisect
import array
import collections
import itertools
import CGAT.GTF as GTF
import CGAT.Bed as Bed
import CGAT.IOTools as IOTools
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta
import CGAT.Stats as Stats
import CGAT.SequenceProperties as SequenceProperties
import CGAT.Genomics as Genomics
import CGAT.Intervals as Intervals

import bx
import bx.bbi.bigwig_file
import bx.intervals.io
import bx.intervals.intersection
import alignlib
import numpy
import CGAT.IndexedGenome as IndexedGenome
import pysam


class Counter:
    """
    This class does not remove small exons/introns,
    so beware ENSEMBL that stores frameshifts as mini-exons.
    """

    header = [ "contig", "strand" ]

    mMinIntronSize = 10

    def __init__(self, fasta = None, section = None, options = None, prefix = None):

        self.fasta = fasta
        self.section = section
        self.options = options
        
        if prefix:
            self.header = tuple(["%s%s" % (prefix, x) for x in self.header ])

        # if true, entry is skipped
        self.skip = False

        # counter 
        self.counter = E.Counter()

    def __call__(self, gffs):
        self.mGFFs = gffs
        self.skip = False
        self.count()

    def __str__(self):
        return "\t".join( (self.contig, self.strand) )

    def getHeader(self):
        if self.section:
            return "\t".join( ["%s_%s" % (self.section, x) for x in self.header] )
        else:
            return "\t".join( self.header )

    def count(self):
        self.contig = self.getContig()
        self.strand = self.getStrand()

    def getContig(self):
        contig = self.mGFFs[0].contig
        if self.fasta: contig = self.fasta.getToken( contig)
        return contig

    def getStrand(self):
        return self.mGFFs[0].strand

    def getGeneId( self ):
        return self.mGFFs[0].gene_id
    
    def getSequence(self, segments):
        """get sequence from a set of segments."""

        contig = self.getContig()
        strand = self.getStrand()
        
        s = []
        for start,end in segments:
            s.append(self.fasta.getSequence( contig, strand, start, end ))
            
        if Genomics.IsNegativeStrand( strand ):
            return( string.translate( "".join(s), string.maketrans("ACGTacgt", "TGCAtgca") )[::-1] )
        else:
            return "".join(s)

    def getExons( self ):
        """merge small introns into single exons. The following features are aggregated
        as exons: exon, CDS, UTR, UTR3, UTR5
        """
        ranges = GTF.asRanges( self.mGFFs, feature = ("exon", "CDS", "UTR", "UTR5", "UTR3" ) )
        assert len(ranges) > 0, "no exons in gene"
        return Intervals.combineAtDistance( ranges,
                                            self.mMinIntronSize )

    def getCDS( self ):
        """merge small introns into single exons. The following features are aggregated
        as exons: exon, CDS, UTR, UTR3, UTR5
        """
        ranges = GTF.asRanges( self.mGFFs, feature = ( "CDS", ) )
        return Intervals.combineAtDistance( ranges,
                                            self.mMinIntronSize )

    def getIntrons( self ):
        exons = self.getExons()
        assert len(exons) > 0, "no exons in gene"
        introns = []
        last = exons[0][1]
        for e in exons[1:]:
            introns.append( (last, e[0] ) )
            last = e[1]
        return introns

    def getSegments( self ):
        if self.section == "exons":
            return self.getExons()
        elif self.section == "introns":
            return self.getIntrons()
        else:
            return self.getExons()

    def getUTRs( self ):
        '''return a tuple with 5' and 3' UTRs.'''
        exons = self.getExons()
        cds = self.getCDS()
        utr3, utr5 = [], []
        if len(cds) == 0: return utr5, utr3
        strand = self.getStrand()
        cds_start, cds_end = cds[0][0], cds[-1][1]
        midpoint = ( cds_end - cds_start ) / 2 + cds_start
        for start, end in Intervals.truncate( exons, cds ):
            if end - start > 3:
                if start < midpoint:
                    utr5.append( (start,end) )
                else:
                    utr3.append( (start,end) )
        if strand == "-":
            utr5, utr3 = utr3, utr5
        return utr5, utr3

cimport numpy
DTYPE_INT = numpy.int
ctypedef numpy.int_t DTYPE_INT_t
DTYPE_FLOAT = numpy.float
ctypedef numpy.float_t DTYPE_FLOAT_t


cdef __add( numpy.ndarray[DTYPE_INT_t, ndim=1]counts, positions, int offset, int length ):
    '''add positions to counts vector'''
    cdef int p, pos
    for p in positions:
        pos = p - offset
        for pos from 0 <= pos < length:
            counts[ pos ] += 1

##-----------------------------------------------------------------------------------
class CounterReadCoverage(Counter):
    '''compute read coverage for all exons in a transcript. 

    Requires bam files to compute that coverage. Multiple bam
    files can be supplied, these will be summed up.

    Counts are separated into sense, antisense and any sense.
    '''
    
    header = ("length",) +\
        tuple( [ "%s_%s" % (x,y) for x,y in itertools.product( ("sense", "antisense", "anysense"),
                                                               ( ("pcovered", "nreads", ) + Stats.Summary().getHeaders() )) ] )
               
    # discard segments with size > mMaxLength in order
    # to avoid out-of-memory
    mMaxLength = 100000

    def __init__(self, bamfiles, *args, **kwargs ):
        Counter.__init__(self, *args, **kwargs )
        if not bamfiles: raise ValueError("supply --bam-file options for readcoverage")
        self.mBamFiles = bamfiles

    def count(self):
        
        segments = self.getSegments()

        cdef AlignedRead read
        cdef int length = sum( [x[1] - x[0] for x in segments ] )
        cdef numpy.ndarray[DTYPE_INT_t, ndim=1] counts_sense = numpy.zeros( length, dtype = numpy.int )
        cdef numpy.ndarray[DTYPE_INT_t, ndim=1] counts_antisense = numpy.zeros( length, dtype = numpy.int )
        cdef int p, pos, offset

        reads_sense, reads_antisense = set(), set()

        contig = self.getContig()
        if self.getStrand() == "+":
            is_reverse = False
        else:
            is_reverse = True
 
        l = 0
        for start, end in segments:
            if end - start > self.mMaxLength: return []
            offset = start - l
            for samfile in self.mBamFiles:
                for read in samfile.fetch( contig, start, end ):
                    # only count positions actually overlapping
                    positions = read.positions
                    if not positions: continue
                    if is_reverse == read.is_reverse:
                        for p in positions:
                            pos = p - offset
                            if 0 <= pos < length:
                                counts_sense[ pos ] += 1
                        reads_sense.add( read.qname )
                    else:
                        for p in positions:
                            pos = p - offset
                            if 0 <= pos < length:
                                counts_antisense[ pos ] += 1
                        reads_antisense.add( read.qname )

            l += end - start

        self.length = length
        self.nreads_sense = len(reads_sense)
        self.nreads_antisense = len(reads_antisense)
        self.nreads_anysense = self.nreads_sense + self.nreads_antisense

        counts_anysense = counts_sense + counts_antisense

        counts_sense = counts_sense[ counts_sense > 0]
        counts_antisense = counts_antisense[ counts_antisense > 0]
        counts_anysense = counts_anysense[ counts_anysense > 0]

        self.counts_sense = counts_sense
        self.counts_antisense = counts_antisense
        self.counts_anysense = counts_anysense
        
    def __str__(self):

        r = [ "%i" % self.length ]
        
        for direction, counts, nreads in zip ( ("sense", "antisense", "anysense"),
                                               (self.counts_sense, self.counts_antisense, self.counts_anysense),
                                               (self.nreads_sense, self.nreads_antisense, self.nreads_anysense) ):
            r.append( "%5.2f" % (100.0 * len(counts) / self.length) )
            r.append( "%i" % (nreads) )
            r.append( str( Stats.Summary( counts, mode = "int" ) ) )

        return "\t".join( r )

##-----------------------------------------------------------------------------------
class CounterReadCounts(Counter):
    '''compute number of reads overlapping with exons.

    Requires bam files to compute that coverage. Multiple bam
    files can be supplied, these will be summed up.

    Both unique and non-unique counts are collected. Uniqueness
    is simply checked through alignment start position.

    Counts are separated into sense, antisense and any sense.

    If ``weight_multi_mapping`` is set, counts are weigthed by the NH flag.
    '''
    
    header = ( [ "%s_%s" % (x,y) for x,y in itertools.product( ( "sense", "antisense", "anysense"),
                                                                ( "unique_counts", "all_counts") ) ] )
    
    def __init__(self, bamfiles, *args, weight_multi_mapping = False, **kwargs ):
        Counter.__init__(self, *args, **kwargs )
        if not bamfiles: raise ValueError("supply --bam-file options for readcoverage")
        self.mBamFiles = bamfiles
        self.weight_multi_mapping = weight_multi_mapping

    def count(self):
        '''count reads.'''
        # For performance reasons, two separate counting implementations.
        if self.weight_multi_mapping:
            self.countFloat()
        else:
            self.countInteger()

    def countFloat(self):
        '''count by weighting multi-mapping reads.'''

        
        segments = self.getSegments()
        contig = self.getContig()

        cdef float nsense_unique_counts = 0
        cdef float nsense_all_counts = 0
        cdef float nantisense_unique_counts = 0
        cdef float nantisense_all_counts = 0
        cdef float nanysense_unique_counts = 0
        cdef float nanysense_all_counts = 0
        cdef float weight = 0
        cdef int nh = 0
        cdef long last_any_pos = -1 
        cdef long last_sense_pos = -1
        cdef long last_anti_pos = -1

        if self.getStrand() == "+":
            is_reverse = False
        else:
            is_reverse = True

        for start, end in segments:
            for samfile in self.mBamFiles:
                last_any_pos = -1
                last_sense_pos = -1
                last_anti_pos = -1
                for read in samfile.fetch( contig, start, end ):
                    if not read.overlap( start, end ): continue
                    try:
                        nh = read.opt( 'NH' )
                    except KeyError:
                        nh = 1
                    
                    weight = 1.0/nh
                        
                    nanysense_all_counts += weight
                    if last_any_pos != read.pos:
                        last_any_pos = read.pos
                        nanysense_unique_counts += weight
                        
                    if is_reverse == read.is_reverse:
                        nsense_all_counts += weight
                        if last_sense_pos != read.pos:
                            last_sense_pos = read.pos
                            nsense_unique_counts += weight
                    else:
                        nantisense_all_counts += weight
                        if last_anti_pos != read.pos:
                            last_anti_pos = read.pos
                            nantisense_unique_counts += weight

        self.result = (nsense_unique_counts,
                       nsense_all_counts,
                       nantisense_unique_counts,
                       nantisense_all_counts,
                       nanysense_unique_counts,
                       nanysense_all_counts )

    def countInteger(self):
        '''count all reads equally.'''

        segments = self.getSegments()
        contig = self.getContig()

        cdef int nsense_unique_counts = 0
        cdef int nsense_all_counts = 0
        cdef int nantisense_unique_counts = 0
        cdef int nantisense_all_counts = 0
        cdef int nanysense_unique_counts = 0
        cdef int nanysense_all_counts = 0
        cdef long last_any_pos = -1 
        cdef long last_sense_pos = -1
        cdef long last_anti_pos = -1

        if self.getStrand() == "+":
            is_reverse = False
        else:
            is_reverse = True

        for start, end in segments:
            for samfile in self.mBamFiles:
                last_any_pos = -1
                last_sense_pos = -1
                last_anti_pos = -1
                for read in samfile.fetch( contig, start, end ):
                    if not read.overlap( start, end ): continue

                    nanysense_all_counts += 1
                    if last_any_pos != read.pos:
                        last_any_pos = read.pos
                        nanysense_unique_counts += 1
                        
                    if is_reverse == read.is_reverse:
                        nsense_all_counts += 1
                        if last_sense_pos != read.pos:
                            last_sense_pos = read.pos
                            nsense_unique_counts += 1
                    else:
                        nantisense_all_counts += 1
                        if last_anti_pos != read.pos:
                            last_anti_pos = read.pos
                            nantisense_unique_counts += 1

        self.result = (nsense_unique_counts,
                       nsense_all_counts,
                       nantisense_unique_counts,
                       nantisense_all_counts,
                       nanysense_unique_counts,
                       nanysense_all_counts )

    def __str__(self):
        return "\t".join( map(str, (self.result)))



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
fasta2distances.py - analyze pairs of sequences
=============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script computes various distances between sequences.

Usage
-----

Example::

   python fasta2distances.py --help

Type::

   python fasta2distances.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import tempfile
import subprocess
import optparse
import math

"""
"""

import CGAT.Experiment as E
import CGAT.Genomics as Genomics
import numpy

import CGAT.FastaIterator as FastaIterator
import Bio.Alphabet.IUPAC

##-------------------------------------------------------------------------
def FilterAlignedPairForPositions( seq1, seq2, method ):
    """given the method, return a set of aligned sequences
    only containing certain positions.

    Available filters:
    all:        do nothing.
    codon1,codon2,codon3: return 1st, 2nd, 3rd codon positions only.
    d4: only changes within fourfold-degenerate sites
    """

    l1 = len(seq1)
    l2 = len(seq2)

    if method == "all":
        return seq1, seq2
    elif method == "codon1":
        return ("".join([seq1[x] for x in range(0, l1, 3)]), 
                "".join([seq2[x] for x in range(0, l2, 3)]) )
    elif method == "codon2":
        return ("".join([seq1[x] for x in range(1, l1, 3)]), 
                "".join([seq2[x] for x in range(1, l2, 3)]) )
    elif method == "codon3":
        return ("".join([seq1[x] for x in range(2, l1, 3)]), 
                "".join([seq2[x] for x in range(2, l2, 3)]) )
    elif method == "d4":
        s1 = []
        s2 = []
        for x in range(0, l1, 3):
            codon1 = seq1[x:x+3]
            codon2 = seq2[x:x+3]
            try:
                aa1, deg11, deg12, deg13 = Genomics.GetDegeneracy( codon1 )
                aa2, deg11, deg22, deg23 = Genomics.GetDegeneracy( codon2 )
            except KeyError:
                continue
            if aa1 == aa2 and deg13 == 4 and deg23 == 4:
                s1.append(codon1[2])
                s2.append(codon2[2])                
        return "".join(s1), "".join(s2)


##-----------------------------------------------------------------------------------------------
def CalculateDistanceJC69( info, do_gamma = False, alpha = None ):
    """return Jukes-Cantor distance.
    """
    try:
        p = float(info.mNDifferent) / info.mNAligned
        
        if do_gamma:
            ## not done yet
            distance = 0.75 * alpha * (pow(1 - 4*p/3, -1/ alpha) - 1)
            variance = p*(1 - p)/(pow(1 - 4*p/3, -2/(alpha + 1)) * L)
        else:
            distance = -0.75 * math.log(1.0 - 4.0 * p / 3.0);
            variance = p * ( 1.0 - p) / ( math.pow(1.0 - 4.0 * p / 3, 2.0) * info.mNAligned);
    except:
        raise ValueError
    
    return distance, variance

##-----------------------------------------------------------------------------------------------
def CalculateDistanceT92( info ):
    """
    P,Q: transitions, transversions frequencies
    q: G+C content
    
    d = -2q(1 - q)loge(1 - P/[2q(1 - q)] - Q) -[1 -2q(1 -q)]loge(1 - 2Q)/2,(4.18)
    V(d) = [c12P + c32Q - (c1P + c3Q)2]/n,(4.19)
    where c1 = 1/(1 - P/[2q(1 - q)] - Q), c2 = 1/(1 - 2Q), c3 = 2q(1 - q)(c1 - c2) + c2, and q is the G+C content

    Note: result is undefined if
        the number of transversions is >= 0.5
        the G+C content is 0

    raises ValueErrors for undefined results
    """
    gc = info.getGCContent()

    # if there are no GC or no AT pairs: result is undefined
    if gc == 0 or gc == 1:
        raise ValueError
    
    wg = 2.0 * gc * (1.0 - gc)

    P = float(info.mNTransitions) / info.mNAligned
    Q = float(info.mNTransversions) / info.mNAligned

    a1 = 1.0 - P / wg - Q;
    if a1 <= 0: raise ValueError
    
    a2 = 1.0 - 2.0 * Q;
    if a2 <= 0: raise ValueError

    # print a1, a2, wg, gc, "p=", P, "q=", Q, str(info)

    distance = -wg * math.log(a1) - 0.5 * (1.0 - wg) * math.log(a2);

    c1 = 1/a1;
    c2 = 1/a2;
    c3 = wg*(c1 - c2) + c2;
    
    variance = (c1 * c1 * P + c3 * c3 * Q - math.pow(c1 * P + c3 * Q, 2.0))/ info.mNAligned;

    return distance, variance

##--------------------------------------------------------------------------------------------
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: fasta2distances.py 2781 2009-09-10 11:33:14Z andreas $")
    
    parser.add_option( "--filters", dest="filters", type="string",
                      help="Filters to use for filtering sequences [all|codon1|codon2|codon3|d4]." )
    parser.add_option( "--fields", dest="fields", type="string",
                      help="Fields to output [aligned|nunaligned1|nunaligned2|identical|transitions|transversions|jc69|t92]." )

    parser.set_defaults(
        filename_map = None,
        filters="all,codon1,codon2,codon3,d4",
        gap_char = "-",
        fields="aligned,unaligned1,unaligned2,identical,transitions,transversions,jc69,t92",
        )
    
    (options, args) = E.Start( parser, add_pipe_options = True)

    options.filters = options.filters.split(",")
    options.fields = options.fields.split(",")    
    
    iterator = FastaIterator.FastaIterator( options.stdin )

    headers = ["id1", "id2"]
    for f in options.filters:
        headers += list( map( lambda x: "%s_%s" % (f, x), options.fields ) )

    options.stdout.write( "\t".join(headers) + "\n")

    while 1:
        try:
            cur_record = iterator.next()
            if cur_record is None: break
            first_record = cur_record
            cur_record = iterator.next()
            if cur_record is None: break
            second_record = cur_record

        except StopIteration:
            break

        if len(first_record.sequence) != len(second_record.sequence):
            raise "sequences %s and %s of unequal length" % (first_record.title, second_record.title)            

        if len(first_record.sequence) % 3 != 0:
            raise "sequence %s not multiple of 3" % first_record.title

        # old: Bio.Alphabet.IUPAC.extended_dna.letters
        alphabet = "ACGT" + options.gap_char

        result = []
        for f in options.filters:
            
            s1, s2= FilterAlignedPairForPositions( first_record.sequence,
                                                   second_record.sequence,
                                                   f )

            info = Genomics.CalculatePairIndices( s1, s2, options.gap_char )

            for field in options.fields:
                
                if field == "aligned":
                    c = "%i" % info.mNAligned
                elif field == "unaligned1":
                    c = "%i" % info.mNUnaligned1
                elif field == "unaligned2":
                    c = "%i" % info.mNUnaligned2
                elif field == "transversions":
                    c = "%i" % info.mNTransversions
                elif field == "transitions":
                    c = "%i" % info.mNTransitions
                elif field == "identical":
                    c = "%i" % info.mNIdentical
                elif field == "jc69":
                    try:
                        c = "%6.4f" % CalculateDistanceJC69( info )[0]
                    except ValueError:
                        c = "nan"
                elif field == "t92":
                    try:
                        c = "%6.4f" % CalculateDistanceT92( info )[0]
                    except ValueError:
                        c = "nan"
                else:
                    raise "Unknown field %s" % field

                result.append( c )
                
        options.stdout.write( "%s\t%s\t%s\n" % (first_record.title,
                                                second_record.title,
                                                "\t".join(result) ))
        
    E.Stop()



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
bed2stats.py - summary of bed file contents
============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script takes a bed-formatted file as input and outputs the number
of intervals and bases in the bed file. Counts can be computed per
contig or per track in the bed file.

Usage
-----

Example::

   python bed2table.py --help

Type::

   python bed2table.py --help

for command line help.

Documentation
-------------

Code
----

'''
import os
import sys
import string
import re
import collections
import CGAT.Bed as Bed
import CGAT.Experiment as E
import CGAT.IndexedFasta as IndexedFasta

class Counter:

    headers = ["ncontigs", "nintervals", "nbases" ]

    def __init__(self):
        self.intervals_per_contig = collections.defaultdict(int)
        self.bases_per_contig = collections.defaultdict(int)
        
    def add( self, bed ):
        self.intervals_per_contig[bed.contig] += 1
        self.bases_per_contig[bed.contig] += bed.end - bed.start

    def __str__(self):
        return "%i\t%i\t%i" % (len( self.intervals_per_contig),
                               sum( self.intervals_per_contig.values()),
                               sum( self.bases_per_contig.values())) 


##------------------------------------------------------------
def main( argv = None ):

    if argv == None: argv = sys.argv

    parser = E.OptionParser( version = "%prog version: $Id: gtf2table.py 2888 2010-04-07 08:48:36Z andreas $", usage = globals()["__doc__"])

    parser.add_option("-g", "--genome-file", dest="genome_file", type="string",
                      help="filename with genome [default=%default]."  )

    parser.add_option("-n", "--per-name", dest="per_name", action="store_true",
                      help="compute counts per name [default=%default]."  )

    parser.add_option("-c", "--per-contig", dest="per_contig", action="store_true",
                      help="compute counts per contig [default=%default]."  )

    parser.add_option("-t", "--per-track", dest="per_track", action="store_true",
                      help="compute counts per track [default=%default]."  )

    parser.set_defaults(
        genome_file = None,
        per_name = False,
        per_track = False,
        )

    (options, args) = E.Start( parser, argv )

    # get files
    if options.genome_file:
        fasta = IndexedFasta.IndexedFasta( options.genome_file )
    else:
        fasta = None

    counts = collections.defaultdict( Counter )

    if options.per_track:
        keyf = lambda x: x.track
    elif options.per_name:
        keyf = lambda x: x.name
    elif options.per_contig:
        keyf = lambda x: x.contig
    else:
        keyf = lambda x: "all"

    for bed in Bed.iterator(options.stdin):
        counts[keyf(bed)].add( bed )

    outf = options.stdout

    key = "track"
    outf.write( "%s\t%s\n" % (key, "\t".join( Counter.headers) ))

    for key, count in counts.iteritems():
        outf.write( "%s\t%s\n" % ( key, str(count)) )
        
    E.Stop()

if __name__ == '__main__':
    sys.exit(main(sys.argv))



################################################################################
#   Gene prediction pipeline 
#
#   $Id: gff2chunks.py 2781 2009-09-10 11:33:14Z andreas $
#
#   Copyright (C) 2004 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
gff2chunks.py - split a gff file into chunks
=============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This scripts splits a gff file into chunks. 
The input file needs to be sorted appropriately.

The gff file is split into chunks ensuring that 
overlapping intervals are part of the same chunk

Usage
-----

Example::

   python gff2chunk.py < in.gff 

Type::

   python gff2chunk.py --help

for command line help.

Documentation
-------------

Command line options
--------------------

'''

import sys
import re
import os
import CGAT.Experiment as E

class OutputChunk:
    def __init__(self, options):
        self.mNChunk = 0
        self.mOptions = options

    def createOpen( self, mode = "w" , header = None):
        """open file. Check first, if directory exists.
        """

        self.mNChunk += 1
        filename = self.mOptions.output_pattern % self.mNChunk

        if self.mOptions.dry_run:
            self.mOptions.stdlog.write("# opening file %s\n" % filename )
            return open("/dev/null", mode)

        if mode in ("w", "a"):
            dirname = os.path.dirname(filename)
            if dirname and not os.path.exists( dirname ):
                os.makedirs( dirname )

        if os.path.exists( filename ):
            existed = True
        else:
            existed = False
        
        f = open( filename, mode )

        if header and not existed:
            f.write( header + "\n" )

        return f

    def __call__( self, chunk ):
        """output a chunk into a new file."""
        outfile = self.createOpen()
        for c in chunk: outfile.write( str(c) + "\n" )
        outfile.close()
        return len(chunk)

##------------------------------------------------------------------------
if __name__ == "__main__":

    parser = E.OptionParser( version = "%prog version: $Id: gff2chunks.py 2781 2009-09-10 11:33:14Z andreas $", usage = globals()["__doc__"] )

    parser.add_option( "-p", "--output-pattern", dest="output_pattern", type="string",
                       help="output pattern for filenames. Should contain a '%i' [default=%default]." )

    parser.add_option( "-i", "--min-chunk-size", dest="min_chunk_size", type="int",
                       help="minimum chunk size [default=%default]." )

    parser.add_option( "-n", "--dry-run", dest="dry_run", action="store_true",
                       help="do not create any files [default=%default]." )

    parser.set_defaults(
        method = "overlap",
        dry_run = False,
        output_pattern = "%06i.chunk",
        min_chunk_size = 1,
        )

    (options, args) = E.Start( parser )

    gffs = GTF.iterator( sys.stdin )
    
    ninput, noutput, nchunks = 0, 0, 0

    outputChunk = OutputChunk( options )

    if options.method == "overlap":
        
        last_contig, last_to = None, None
        chunk = []
        for gff in gffs:
            ninput += 1
            if len(chunk) > options.min_chunk_size and \
                    (gff.contig != last_contig or \
                         gff.start > last_to):
                noutput += outputChunk( chunk )
                nchunks += 1
                chunk = []
                last_contig, last_to = gff.contig, gff.end
            
            chunk.append( gff )
            last_to = max( gff.end, last_to )
            
        noutput += outputChunk( chunk )
        nchunks += 1
        
    if options.loglevel >= 1:
        options.stdlog.write( "# ninput=%i, noutput=%i, nchunks=%i\n" % (ninput, noutput, nchunks ) )

    E.Stop()



################################################################################
#

#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Experiment.py - Tools for scripts
=================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

The :mod:`Experiment` modules contains utility functions for logging
and record keeping of scripts.

This module is imported by most CGAT scripts. It provides convenient
and consistent methods for

   * `Record keeping`_
   * `Logging`_
   * `Benchmarking`_

See :doc:`../scripts/script_template` on how to use this module.

The basic usage of this module within a script is::

    """script_name.py - my script
    
    Mode Documentation
    """
    import sys
    import optparse
    import CGAT.Experiment as E

    def main( argv = None ):
        """script main.

        parses command line options in sys.argv, unless *argv* is given.
        """

        if not argv: argv = sys.argv

        # setup command line parser
        parser = E.OptionParser( version = "%prog version: $Id$", 
                                        usage = globals()["__doc__"] )
                                        
        parser.add_option("-t", "--test", dest="test", type="string",
                          help="supply help"  )

        ## add common options (-h/--help, ...) and parse command line 
        (options, args) = E.Start( parser )

        # do something
        # ...
        E.info( "an information message" )
        E.warn( "a warning message")

        ## write footer and output benchmark information.
        E.Stop()

    if __name__ == "__main__":
        sys.exit( main( sys.argv) )


Record keeping
--------------

The central functions in this module are the :py:func:`Start` and
:py:func:`Stop` methods which are called before or after any work is done 
within a script.

.. autofunction:: Experiment.Start

The :py:func:`Start` is called with an E.OptionParser object. 
:py:func:`Start` will add additional command line arguments, such as
``--help`` for command line help or ``--verbose`` to control the :term:`loglevel`.
It can also add optional arguments for scripts needing database access,
writing to multiple output files, etc. 

:py:func:`Start` will write record keeping information to a logfile. Typically, logging
information is output on stdout, prefixed by a `#`, but it can be re-directed to 
a separate file. Below is a typical output::

    # output generated by /ifs/devel/andreas/cgat/beds2beds.py --force --exclusive --method=unmerged-combinations --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz Rela-030m-R1.bed.gz
    # job started at Thu Mar 29 13:06:33 2012 on cgat150.anat.ox.ac.uk -- e1c16e80-03a1-4023-9417-f3e44e33bdcd
    # pid: 16649, system: Linux 2.6.32-220.7.1.el6.x86_64 #1 SMP Fri Feb 10 15:22:22 EST 2012 x86_64
    # exclusive                               : True
    # filename_update                         : None
    # ignore_strand                           : False
    # loglevel                                : 1
    # method                                  : unmerged-combinations
    # output_filename_pattern                 : 030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
    # output_force                            : True
    # pattern_id                              : (.*).bed.gz
    # stderr                                  : <open file \'<stderr>\', mode \'w\' at 0x2ba70e0c2270>
    # stdin                                   : <open file \'<stdin>\', mode \'r\' at 0x2ba70e0c2150>
    # stdlog                                  : <open file \'030m.intersection.tsv.log\', mode \'a\' at 0x1f1a810>
    # stdout                                  : <open file \'<stdout>\', mode \'w\' at 0x2ba70e0c21e0>
    # timeit_file                             : None
    # timeit_header                           : None
    # timeit_name                             : all
    # tracks                                  : None

The header contains information about:

    * the script name (``beds2beds.py``)
    * the command line options (``--force --exclusive --method=unmerged-combinations --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz Rela-030m-R1.bed.gz``)
    * the time when the job was started (``Thu Mar 29 13:06:33 2012``)
    * the location it was executed (``cgat150.anat.ox.ac.uk``)
    * a unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)
    * the pid of the job (``16649``)
    * the system specification (``Linux 2.6.32-220.7.1.el6.x86_64 #1 SMP Fri Feb 10 15:22:22 EST 2012 x86_64``)

It is followed by a list of all options that have been set in the script.
    
Once completed, a script will call the :py:func:`Stop` function to signify the end of the experiment.
 
.. autofunction:: Experiment.Stop

:py:func:`Stop` will output to the log file that the script has concluded successfully. Below is typical output::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36  0.45  0.00  0.01 -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

The footer contains information about:

   * the job has finished (``job finished``)
   * the time it took to execute (``11 seconds``)
   * when it completed (``Thu Mar 29 13:06:44 2012``)
   * some benchmarking information (``11.36  0.45  0.00  0.01``) which is 
         ``user time``, ``system time``, ``child user time``, ``child system time``.
   * the unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)

The unique job id can be used to easily retrieve matching information from a concatenation of 
log files.

Benchmarking
------------




Complete reference
------------------
'''

import string,re,sys,time,inspect,os,logging
import collections, types, subprocess, gzip

import optparse
import textwrap

class DefaultOptions:
    stdlog = sys.stdout
    stdout = sys.stdout
    stderr = sys.stderr
    stdin = sys.stdin
    loglevel = 2
    timeit_file = None

global_starting_time = time.time()
global_options = DefaultOptions()
global_args    = None
# import hashlib
# global_id = hashlib.md5(time.asctime(time.localtime(time.time()))).hexdigest()
import uuid
global_id      = uuid.uuid4()
global_benchmark = collections.defaultdict( int )

###################################################################################
## The code for BetterFormatter has been taken from
## http://code.google.com/p/yjl/source/browse/Python/snippet/BetterFormatter.py
__copyright__ = """
Copyright (c) 2001-2006 Gregory P. Ward.  All rights reserved.
Copyright (c) 2002-2006 Python Software Foundation.  All rights reserved.
Copyright (c) 2011 Yu-Jie Lin.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  * Neither the name of the author nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

class BetterFormatter(optparse.IndentedHelpFormatter):

  def __init__(self, *args, **kwargs):

    optparse.IndentedHelpFormatter.__init__(self, *args, **kwargs)
    self.wrapper = textwrap.TextWrapper(width=self.width)

  def _formatter(self, text):

    return '\n'.join(['\n'.join(p) for p in map(self.wrapper.wrap,
        self.parser.expand_prog_name(text).split('\n'))])

  def format_description(self, description):

    if description:
      return self._formatter(description) + '\n'
    else:
      return ''

  def format_epilog(self, epilog):

    if epilog:
      return '\n' + self._formatter(epilog) + '\n'
    else:
      return ''

  def format_usage(self, usage):

    return self._formatter(optparse._("Usage: %s\n") % usage)

  def format_option(self, option):
    # Ripped and modified from Python 2.6's optparse's HelpFormatter
    result = []
    opts = self.option_strings[option]
    opt_width = self.help_position - self.current_indent - 2
    if len(opts) > opt_width:
      opts = "%*s%s\n" % (self.current_indent, "", opts)
      indent_first = self.help_position
    else:                       # start help on same line as opts
      opts = "%*s%-*s  " % (self.current_indent, "", opt_width, opts)
      indent_first = 0
    result.append(opts)
    if option.help:
      help_text = self.expand_default(option)
      # Added expand program name
      help_text = self.parser.expand_prog_name(help_text)
      # Modified the generation of help_line
      help_lines = []
      wrapper = textwrap.TextWrapper(width=self.help_width)
      for p in map(wrapper.wrap, help_text.split('\n')):
        if p:
          help_lines.extend(p)
        else:
          help_lines.append('')
      # End of modification
      result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
      result.extend(["%*s%s\n" % (self.help_position, "", line)
                     for line in help_lines[1:]])
    elif opts[-1] != "\n":
      result.append("\n")
    return "".join(result)


## End of BetterFormatter()
#################################################################
#################################################################
#################################################################
class AppendCommaOption(optparse.Option):
    '''Option with additional parsing capabilities.

    * "," in arguments to options that have the action 'append' 
      are treated as a list of options. This is what galaxy does, 
      but generally convenient.

    * Option values of "None" and "" are treated as default values.
    
    
    '''
#    def check_value( self, opt, value ):
#        # do not check type for ',' separated lists
#        if "," in value: 
#            return value
#        else:
#            return optparse.Option.check_value( self, opt, value )
#
#    def take_action(self, action, dest, opt, value, values, parser):
#        if action == "append" and "," in value:
#            lvalue = value.split(",")
#            values.ensure_value(dest, []).extend(lvalue)
#        else:
#            optparse.Option.take_action(
#                self, action, dest, opt, value, values, parser)
#

    def convert_value( self, opt, value ):
        if value is not None:
            if self.nargs == 1:
                if self.action == "append":
                    if "," in value:
                        return [self.check_value(opt, v ) for v in value.split(",") if v != ""]
                    else:
                        if value != "":
                            return self.check_value( opt, value )
                        else:
                            return value
                else:
                    return self.check_value( opt, value )
            else:
                return tuple( [self.check_value(opt, v) for v in value] )

    # why is it necessary to pass action and dest to this function when
    # they could be accessed as self.action and self.dest? 
    def take_action( self, action, dest, opt, value, values, parser):

        if action == "append" and type( value ) == list: 
            values.ensure_value(dest, []).extend( value )
        else:
            optparse.Option.take_action(
                self, action, dest, opt, value, values, parser )

#################################################################
#################################################################
#################################################################
class OptionParser( optparse.OptionParser ):
    '''CGAT derivative of OptionParser.

    '''
    
    def __init__(self, *args, **kwargs):
        optparse.OptionParser.__init__(self, *args, 
                                       option_class = AppendCommaOption,
                                       formatter = BetterFormatter(),
                                       **kwargs)

        # set new option parser
        # parser.formatter = BetterFormatter()
        # parser.formatter.set_parser(parser)


#################################################################
def openFile( filename, mode = "r", create_dir = False ):
    '''open file in *filename* with mode *mode*.

    If *create* is set, the directory containing filename
    will be created if it does not exist.

    gzip - compressed files are recognized by the
    suffix ``.gz`` and opened transparently.

    Note that there are differences in the file
    like objects returned, for example in the
    ability to seek.

    returns a file or file-like object.
    '''
    
    _, ext = os.path.splitext( filename )

    if create_dir:
        dirname = os.path.dirname( filename )
        if dirname and not os.path.exists( dirname ):
            os.makedirs( dirname )
            
    if ext.lower() in (".gz", ".z"):
        return gzip.open( filename, mode )
    else:
        return open( filename, mode )

#################################################################
def getHeader():
    """return a header string with command line options and timestamp

    """
    system, host, release, version, machine = os.uname()
    return "# output generated by %s\n# job started at %s on %s -- %s\n# pid: %i, system: %s %s %s %s" %\
           (" ".join(sys.argv),
            time.asctime(time.localtime(time.time())),
            host,
            global_id,
            os.getpid(),
            system, release, version, machine)

def getParams( options = None):
    """return a string containing script parameters.

    Parameters are all variables that start with ``param_``.
    """
    result = []
    if options:
        members = options.__dict__
        for k, v in sorted(members.items()):
            result.append("# %-40s: %s" % (k, str(v).encode("string_escape")))
    else:
        vars = inspect.currentframe().f_back.f_locals
        for var in filter(lambda x: re.match("param_",x), vars.keys()):
            result.append("# %-40s: %s" % (var, str(vars[var]).encode("string_escape")))

    if result:
        return "\n".join(result)
    else:
        return "# no parameters."

def getFooter():
    """return a header string with command line options and
    timestamp.
    """
    return "# job finished in %i seconds at %s -- %s -- %s" %\
           (time.time() - global_starting_time,
            time.asctime(time.localtime(time.time())),
            " ".join( map( lambda x: "%5.2f" % x, os.times()[:4])),
            global_id)

######################################################################
## Deprecated old call interface
def GetFooter(): return getFooter()
def GetHeader(): return getHeader()
def GetParams(): return getParams()

######################################################################
def Start( parser = None,
           argv = sys.argv,
           quiet = False,
           no_parsing = False,
           add_csv_options = False,
           add_mysql_options = False,
           add_psql_options = False,
           add_pipe_options = True,
           add_cluster_options = False,
           add_output_options = False,
           return_parser = False):
    """set up an experiment.

    *param parser* an :py:class:`E.OptionParser` instance with commandi line options.
    *param argv* command line options to parse. Defaults to :py:data:`sys.argv`
    *quiet* set :term:`loglevel` to 0 - no logging
    *no_parsing* do not parse command line options
    *return_parser* return the parser object, no parsing
    *add_csv_options* add common options for parsing :term:`tsv` separated files
    *add_mysql_options* add common options for connecting to mysql_ databases
    *add_psql_options* add common options for connecting to postgres_ databases
    *add_pipe_options* add common options for redirecting input/output
    *add_cluster_options* add common options for scripts submitting jobs to the cluster
    *add_output_options* add commond options for working with multiple output files
    *returns* a tuple (options,args) with options (a :py:class:`E.OptionParser` object 
              and a list of positional arguments.

    The :py:func:`Start` method will also set up a file logger.

    The default options added by this method are:

    ``-v/--verbose`` 
        the :term:`loglevel`

    ``timeit``
        turn on benchmarking information and save to file

    ``timeit-name``
         name to use for timing information, 

    ``timeit-header``
         output header for timing information.

    Optional options added are:

    add_csv_options

       ``dialect``
            csv_dialect. the default is ``excel-tab``, defaulting to :term:`tsv` formatted files.
    
    add_psql_options
       ``-C/--connection``
           psql connection string
       ``-U/--user``
           psql user name

    add_cluster_options
       ``--use-cluster``
           use cluster
       ``--cluster-priority``
           cluster priority to request
       ``--cluster-queue``
           cluster queue to use
       ``--cluster-num-jobs``
           number of jobs to submit to the cluster at the same time
       ``--cluster-options``
           additional options to the cluster for each job.

    add_output_options
       ``-P/--output-filename-pattern``
            Pattern to use for output filenames.

    """
    
    if not parser:
        parser = OptionParser( version = "%prog version: $Id: Experiment.py 2803 2009-10-22 13:41:24Z andreas $" )

    global global_options, global_args, global_starting_time

    global_starting_time = time.time()

    parser.add_option("-v", "--verbose", dest="loglevel", type="int",
                      help="loglevel [%default]. The higher, the more output." )

    parser.add_option( "--timeit", dest='timeit_file', type="string",
                       help="store timeing information in file [%default]." )        
    parser.add_option( "--timeit-name", dest='timeit_name', type="string",
                       help="name in timing file for this class of jobs [%default]." )
    parser.add_option( "--timeit-header", dest='timeit_header', action="store_true",
                       help="add header for timing information [%default]." )        
                                     
    if quiet:
        parser.set_defaults( loglevel = 0 )
    else:
        parser.set_defaults( loglevel = 1 )

    parser.set_defaults(
        timeit_file = None,
        timeit_name = 'all',
        timeit_header = None,
        )

    if add_csv_options:
        parser.add_option("--dialect", dest="csv_dialect", type="string",
                          help="csv dialect to use [%default]." )

        parser.set_defaults(
            csv_dialect = "excel-tab",
            csv_lineterminator = "\n",
            )

    if add_psql_options:
        parser.add_option("-C", "--connection", dest="psql_connection", type="string",
                          help="psql connection string [%default]."  )
        parser.add_option("-U", "--user", dest="user", type="string",
                          help="database user name [%default]."  )
        
        parser.set_defaults( psql_connection = "fgu202:postgres" )
        parser.set_defaults( user = "" )
        
    if add_cluster_options:
        parser.add_option( "--use-cluster", dest="use_cluster", action = "store_true",
                          help="use cluster [%default]."  )
        parser.add_option( "--cluster-priority", dest="cluster_priority", type="int",
                          help="set job priority on cluster [%default]."  )
        parser.add_option( "--cluster-queue", dest="cluster_queue", type="string",
                          help="set cluster queue [%default]."  )
        parser.add_option( "--cluster-num-jobs", dest="cluster_num_jobs", type="int",
                          help="number of jobs to submit to the queue execute in parallel [%default]."  )
        parser.add_option( "--cluster-options", dest="cluster_options", type="string",
                          help="additional options for cluster jobs, passed on to qrsh [%default]."  )
        
        parser.set_defaults( use_cluster = False,
                             cluster_queue = "all.q",
                             cluster_priority = -10,
                             cluster_num_jobs = 100,
                             cluster_options = "")

    if add_output_options:
        parser.add_option( "-P", "--output-filename-pattern", dest="output_filename_pattern", type="string" ,
                           help="OUTPUT filename pattern for various methods [%default].")

        parser.add_option( "-F", "--force", dest="output_force", action="store_true",
                           help="force over-writing of existing files." )

        parser.set_defaults( output_filename_pattern = "%s",
                             output_force = False )

    if add_pipe_options:
        parser.add_option("-I", "--stdin", dest="stdin", type="string",
                          help="file to read stdin from [default = stdin].",
                          metavar = "FILE"  )
        parser.add_option("-L", "--log", dest="stdlog", type="string",
                          help="file with logging information [default = stdout].",
                          metavar = "FILE"  )
        parser.add_option("-E", "--error", dest="stderr", type="string",
                          help="file with error information [default = stderr].",
                          metavar = "FILE"  )
        parser.add_option("-S", "--stdout", dest="stdout", type="string",
                          help="file where output is to go [default = stdout].",
                          metavar = "FILE"  )
        
        parser.set_defaults( stderr = sys.stderr )
        parser.set_defaults( stdout = sys.stdout )
        parser.set_defaults( stdlog = sys.stdout )
        parser.set_defaults( stdin = sys.stdin )
        
    if add_mysql_options:
        parser.add_option("-H", "--host", dest="host", type="string",
                          help="mysql host [%default]."  )
        parser.add_option("-D", "--database", dest="database", type="string",
                          help="mysql database [%default]."  )
        parser.add_option("-U", "--user", dest="user", type="string",
                          help="mysql username [%default]."  )
        parser.add_option("-P", "--password", dest="password", type="string",
                          help="mysql password [%default]."  )
        parser.add_option("-O", "--port", dest="port", type="int",
                          help="mysql port [%default]."  )
        
        parser.set_defaults( host = "db",
                             port = 3306,
                             user = "",
                             password = "",
                             database = "" )

    if return_parser:
        return parser

    if not no_parsing:
        (global_options, global_args) = parser.parse_args( argv[1:] )

    if add_pipe_options:
        if global_options.stdout != sys.stdout: 
            global_options.stdout = openFile(global_options.stdout, "w")
        if global_options.stderr != sys.stderr:
            if global_options.stderr == "stderr":
                global_options.stderr = global_options.stderr
            else:
                global_options.stderr = openFile(global_options.stderr, "w")
        if global_options.stdlog != sys.stdout:
            global_options.stdlog = openFile(global_options.stdlog, "a")            
        if global_options.stdin != sys.stdin: 
            global_options.stdin = openFile(global_options.stdin, "r")
    else:
        global_options.stderr = sys.stderr
        global_options.stdout = sys.stdout
        global_options.stdlog = sys.stdout
        global_options.stdin = sys.stdin
    
    if global_options.loglevel >= 1:
        global_options.stdlog.write(getHeader() + "\n" )
        global_options.stdlog.write(getParams( global_options) + "\n")
        global_options.stdlog.flush()

    ## configure logging
    ## map from 0-10 to logging scale
    ## 0: quiet
    ## 1: little verbositiy
    ## >1: increased verbosity
    if global_options.loglevel == 0:
        lvl = logging.ERROR
    elif global_options.loglevel == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    if global_options.stdout == global_options.stdlog:
        logging.basicConfig(
            level=lvl,
            format='# %(asctime)s %(levelname)s %(message)s',
            stream = global_options.stdlog )
    else:
        logging.basicConfig(
            level=lvl,
            format='%(asctime)s %(levelname)s %(message)s',
            stream = global_options.stdlog )
        
    return global_options, global_args
    
def Stop():
    """stop the experiment."""

    if global_options.loglevel >= 1 and global_benchmark:
        t = time.time() - global_starting_time
        global_options.stdlog.write( "######### Time spent in benchmarked functions ###################\n" )
        global_options.stdlog.write( "# function\tseconds\tpercent\n" )
        for key, value in global_benchmark.items():
            global_options.stdlog.write( "# %s\t%6i\t%5.2f%%\n" % (key, value, (100.0 * float(value) / t)))
        global_options.stdlog.write( "#################################################################\n" )

    if global_options.loglevel >= 1:
        global_options.stdlog.write(getFooter() + "\n")

    # close files
    if global_options.stdout != sys.stdout: global_options.stdout.close()
    # do not close log, otherwise the following error occurs:
    # Error in sys.exitfunc:
    # Traceback (most recent call last):
    #   File "/net/cpp-group/server/lib/python2.6/atexit.py", line 24, in _run_exitfuncs
    #     func(*targs, **kargs)
    #   File "/net/cpp-group/server/lib/python2.6/logging/__init__.py", line 1472, in shutdown
    #     h.flush()
    #   File "/net/cpp-group/server/lib/python2.6/logging/__init__.py", line 740, in flush
    #     self.stream.flush()
    # ValueError: I/O operation on closed file
    # if global_options.stdlog != sys.stdout: global_options.stdlog.close()
    if global_options.stderr != sys.stderr: global_options.stderr.close()

    if global_options.timeit_file:

        outfile = open(global_options.timeit_file, "a")

        if global_options.timeit_header:
            outfile.write( "\t".join( ("name", "wall", "user", "sys", "cuser", "csys",
                                       "host", "system", "release", "machine",
                                       "start", "end", "path", "cmd" ) ) + "\n" )

        csystem, host, release, version, machine = map(str, os.uname())
        uusr, usys, c_usr, c_sys = map( lambda x: "%5.2f" % x, os.times()[:4])
        t_end = time.time()
        c_wall = "%5.2f" % (t_end - global_starting_time)
        
        if sys.argv[0] == "run.py":
            cmd = global_args[0]
            if len(global_args) > 1:
                cmd += " '" + "' '".join(global_args[1:]) + "'" 
        else:
            cmd = sys.argv[0]

        result = "\t".join( (global_options.timeit_name,
                             c_wall, uusr, usys, c_usr, c_sys,
                             host, csystem, release, machine,
                             time.asctime(time.localtime( global_starting_time )),
                             time.asctime(time.localtime( t_end )),
                             os.path.abspath( os.getcwd() ),
                             cmd ) ) + "\n"

        outfile.write( result )
        outfile.close()

def benchmark(func):
    """decorator collecting wall clock time spent in decorated method."""

    def wrapper(*arg):
        t1 = time.time()
        res = func(*arg)
        t2 = time.time()
        key = "%s:%i" % (func.func_name, func.func_code.co_firstlineno)
        global_benchmark[key] += t2-t1
        global_options.stdlog.write( '## benchmark: %s completed in %6.4f s\n' % (key, (t2-t1)))
        global_options.stdlog.flush()
        return res
    return wrapper

# there are differences whether you cache a function or
# an objects method
def cachedmethod(function):
    '''decorator for caching a method.'''
    return Memoize(function)

class cachedfunction(object):
   """Decorator that caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned, and
   not re-evaluated.

   Taken from http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
   """
   def __init__(self, func):
      self.func = func
      self.cache = {}
   def __call__(self, *args):
      try:
         return self.cache[args]
      except KeyError:
         value = self.func(*args)
         self.cache[args] = value
         return value
      except TypeError:
         # uncachable -- for instance, passing a list as an argument.
         # Better to not cache than to blow up entirely.
         return self.func(*args)
   def __repr__(self):
      """Return the function's docstring."""
      return self.func.__doc__
   def __get__(self, obj, objtype):
      """Support instance methods."""
      return functools.partial(self.__call__, obj)

class Memoize(object):
    def __init__(self, fn):
        self.cache={}
        self.fn=fn

    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    def __call__(self,*args):
        if self.cache.has_key(args):
            return self.cache[args]
        else:
            object = self.cache[args] = self.fn(self.instance, *args)
            return object

def log( loglevel, message ):
    """log message at loglevel."""
    logging.log( loglevel, message )

def info( message ):
    '''log information message, see the :mod:`logging` module'''
    logging.info( message )

def warning( message ):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning( message )

def warn( message ):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning( message )

def debug( message ):
    '''log debugging message, see the :mod:`logging` module'''
    logging.debug( message )

def error( message ):
    '''log error message, see the :mod:`logging` module'''
    logging.error( message )
        
def critical( message):
    '''log critical message, see the :mod:`logging` module'''
    logging.critical( message )

def getOutputFile( section ):
    '''return filename to write to.'''
    return re.sub( "%s", section, global_options.output_filename_pattern )
    
def openOutputFile( section, mode = "w" ):
    """open file for writing substituting section in the
    output_pattern (if defined).

    If the filename ends with ".gz", the output is opened
    as a gzip'ed file.
    """

    fn = getOutputFile( section )
    try:
        if fn == "-":
            return global_options.stdout
        else:
            if not global_options.output_force and os.path.exists( fn ):
                raise OSError( "file %s already exists, use --force to overwrite existing files." % fn )
            if fn.endswith(".gz"):
                return gzip.open( fn, mode )
            else:
                return open( fn, mode )
    except AttributeError:        
        return global_options.stdout


class Counter(object):
    '''a counter class.

    The counter acts both as a dictionary and
    a object permitting attribute access.

    Counts are automatically initialized to 0.
    
    Instantiate and use like this::

       c = Counter()
       c.input += 1
       c.output += 2
       c["skipped"] += 1

       print str(c)
    '''

    __slots__=["_counts"]
    def __init__(self):
        """Store data returned by function."""
        object.__setattr__( self, "_counts", collections.defaultdict( int ) )

    def __setitem__(self, key, value ):
        self._counts[key] = value
    def __getitem__(self, key):
        return self._counts[key]
    def __getattr__(self, name):
        return self._counts[name]
    def __setattr__(self, name, value):
        self._counts[name] = value
    def __str__(self):
        return ", ".join( "%s=%i" % x for x in self._counts.iteritems() )
    def __iadd__(self, other ):
        try:
            for key,val in other.iteritems():
                self._counts[key] += val
        except:
            raise TypeError( "unknown type" )
        return self
    def iteritems(self):
        return self._counts.iteritems()
    def asTable( self ):
        '''return values as tab-separated table (without header).'''
        return '\n'.join( "%s\t%i" % x for x in self._counts.iteritems() )

class Experiment:
    
    mShortOptions = ""
    mLongOptions  = []

    mLogLevel = 0
    mTest = 0
    mDatabaseName = None

    mName = sys.argv[0]


    def __init__(self ):

        # process command-line arguments
        (self.mOptlist, self.mArgs) = self.ParseCommandLine()

        # set options now
        self.ProcessOptions(self.mOptlist)

    def DumpParameters( self ):
        """dump parameters of this object. All parameters start with a lower-case m."""

        members = self.__dict__

	print "#--------------------------------------------------------------------------------------------"
        print "#" + string.join(sys.argv)
        print "# pid: %i, system:" % os.getpid(), string.join(os.uname(), ",") 
	print "#--------------------------------------------------------------------------------------------"
        print "# Parameters for instance of <" + self.mName + "> on " + time.asctime(time.localtime(time.time()))

        member_keys = list(members.keys())
        member_keys.sort()
        for member in member_keys:
            if member[0] == 'm':
                print "# %-40s:" % member, members[member]

	print "#--------------------------------------------------------------------------------------------"
        sys.stdout.flush()

    #-----------------------------> Control functions <--------------------------------

    ##------------------------------------------------------------------------------------
    def ProcessOptions( self, optlist ):
        """Sets options in this module. Please overload as necessary."""

        for o,a in optlist:
            if o in ( "-V", "--Verbose" ):
                self.mLogLevel = string.atoi(a)
            elif o in ( "-T", "--test") :
                self.mTest = 1

    ##------------------------------------------------------------------------------------        
    def ProcessArguments( self, args ):
        """Perform actions as given in command line arguments."""

        if self.mLogLevel >= 1:
            self.DumpParameters()
            
        for arg in args:
            if arg[-1] == ")":
                statement = "self.%s" % arg
            else:
                statement = "self.%s()" % arg
            exec statement

            if self.mLogLevel >= 1:
                print "--------------------------------------------------------------------------------------------"
                print statement + " finished at " + time.asctime(time.localtime(time.time()))
                print "--------------------------------------------------------------------------------------------"

    ##------------------------------------------------------------------------------------            
    def ParseCommandLine( self ):
        """Call subroutine with command line arguments."""

        self.mShortOptions = self.mShortOptions + "V:D:T"
        self.mLongOptions.append( "Verbose=" )
        self.mLongOptions.append( "Database=" )
        self.mLongOptions.append( "Test" )

        try:
            optlist, args = getopt.getopt(sys.argv[1:],
                                          self.mShortOptions,
                                          self.mLongOptions)
        except getopt.error, msg:
            self.PrintUsage()
            print msg
            sys.exit(2)

        return optlist, args

    #--------------------------------------------------------------------------------        
    def Process( self ):
        self.ProcessArguments(self.mArgs)

    #--------------------------------------------------------------------------------
    def PrintUsage( self ):
        """print usage information."""

        print "# valid short options are:", self.mShortOptions
        print "# valid long options are:", str(self.mLongOptions)
    
def run( cmd, return_stdout = False, **kwargs ):
    '''executed a command line cmd.
    
    returns the return code.

    If *return_stdout* is True, the contents of stdout
    are returned.

    ``kwargs`` are passed on to subprocess.call or subprocess.check_output.

    raises OSError if process failed or was terminated.
    '''

    # remove new lines
    cmd = " ".join( re.sub( "\t+", " ", cmd).split( "\n" ) ).strip()

    if "<(" in cmd:
        if "'" in cmd: raise ValueError( "advanced bash syntax combined with single quotes" )
        cmd = """/bin/bash -c '%s'""" % cmd

    if return_stdout:
        return subprocess.check_output( cmd, shell = True, **kwargs )
    else:
        retcode = subprocess.call( cmd, shell=True, **kwargs)
        if retcode < 0:
            raise OSError( "process was terminated by signal %i" % -retcode )
        return retcode




################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
"""
beds2beds.py - decompose bed files
==================================

:Author: Andreas Heger
:Release: $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $
:Date: |today|
:Tags: Python

Purpose
-------

decompose a collection of bed-files into a collection unions/intersections.

merged-combinations
   merge intervals across :term:`bed` files and only report those
   that appear in every file.

unmerged-combinations
   for each :term:`bed` file, report intervals that overlap with intervals
   in every other :term:`bed` file.

If the ``--exclusive`` option is set, report exclusive overlap. Only intervals 
will be reported that overlap in a pairwise comparison but do not overlap with 
intervals in any of the other sets.

This script requires bed files indexed by :term:`tabix`.

Usage
-----

For example, you have ChIP-Seq data for PolII and two transcription factors tf1 and tf2.
The following statement will output four :term:`bed` files::

   python beds2beds.py polii.bed.gz tf1.bed.gz tf2.bed.gz

The four files are containing intervals, that

1. have PolII and tf1 present,
2. have PolII and tf2 present,
3. have tf1 and tf2 present, or
4. have PolII and tf1 and tf2 present.

If the --exclusive option is set, three sets will be output with intervals that

1. have PolII and tf1 present but no tf2,
2. have PolII and tf2 present but no tf1,
3. have tf1 and tf2 present bu no PolII.

Type::

   python beds2beds.py --help

for command line help.

Command line options
--------------------

""" 

import os
import sys
import re
import optparse
import itertools
import collections

import CGAT.Experiment as E
import CGAT.IOTools as IOTools
import CGAT.Bed as Bed
import numpy
import pysam
import CGAT.Intervals as Intervals

def isContainedInAll( contig, start, end, bedfiles ):

    for bedfile in bedfiles:
        try:
            if len(list(bedfile.fetch( contig, start, end ))) == 0:
                return False
        except ValueError:
            return False

    return True

def isContainedInOne( contig, start, end, bedfiles ):

    for bedfile in bedfiles:
        try:
            if len(list(bedfile.fetch( contig, start, end ))) > 0:
                return True
        except ValueError:
            pass

    return False

def combineMergedIntervals( bedfiles ):
    '''combine intervals in a collection of bed files.
    
    Overlapping intervals between tracks are merged.

    Algorithm:

    1. collect all intervals in all tracks into a single track
    2. merge overlapping intervals 
    3. report all intervals that overlap with an interval in each track.

    '''
    
    # get all intervals
    data_per_contig = collections.defaultdict( list )
    for bedfile in bedfiles:
        for contig in bedfile.contigs:
            i = []
            for bed in bedfile.fetch( contig, parser = pysam.asBed() ):
                i.append( (bed.start, bed.end) )
            data_per_contig[contig].extend( i )
    
    # merge intervals
    for contig in data_per_contig.keys():
        data_per_contig[contig] = Intervals.combine( data_per_contig[contig] )

    # filter intervals - take only those present in all bedfiles
    for contig, data in data_per_contig.iteritems():
        for start, end in data:
            if isContainedInAll( contig, start, end, bedfiles ):
                yield contig, start, end

def combineUnmergedIntervals( foreground, background ):
    '''combine intervals in a collection of bed files.
    
    Only intervals in the first track are reported.

    Algorithm:

    1. report all intervals in the first track that overlap with an interval in every other track.
    '''

    intervals = []
    c = 0
    for bed in foreground.fetch( parser = pysam.asBed() ):
        c += 1
        if isContainedInAll( bed.contig, bed.start, bed.end, background ):
            yield bed

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = E.OptionParser( version = "%prog version: $Id: diff_bed.py 2866 2010-03-03 10:18:49Z andreas $", usage = globals()["__doc__"] )

    parser.add_option("-e", "--exclusive", dest="exclusive", action="store_true",
                      help="Intervals reported will be merged across the positive set"
                           " and do not overlap any interval in any of the other sets"
                           " [default=%default]." )

    parser.add_option("-p", "--pattern-id", dest="pattern_id", type="string",
                      help="pattern to convert a filename to an id [default=%default]." )

    parser.add_option( "-m", "--method", dest="method", type="choice",
                       choices = ("merged-combinations", 
                                  "unmerged-combinations" ),
                       help = "method to perform [default=%default]" )
    
    parser.set_defaults(
        pattern_id = "(.*).bed.gz",
        exclusive = False,
        method = "merged-combinations",
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    if len(args) < 2:
        raise ValueError( "at least two arguments required" )

    tags, bedfiles = [], []
    for infile in args:
        bedfiles.append( pysam.Tabixfile( infile, "r" ) )
        tags.append( re.search( options.pattern_id, infile ).groups()[0] )

    indices = range(len(bedfiles))
    is_exclusive = options.exclusive

    if options.method == "merged-combinations":

        if is_exclusive: start = 1    
        else: start = 2

        options.stdout.write( "combination\twithout\tcounts\n" )

        for ncombinants in range( start, len(bedfiles) + 1):
            for combination in itertools.combinations( indices, ncombinants ):
                other = [ x for x in indices if x not in combination ]
                tag = ":".join( [ tags[x] for x in combination ] )
                E.debug( "combination %s started" % tag)
                E.debug( "other: %s" % ":".join( [tags[x] for x in other]) )
                other_bed = [ bedfiles[x] for x in other ]
                outf = IOTools.openFile( E.getOutputFile( tag ), "w", create_dir = True)
                c = E.Counter()
                for contig, start, end in combineMergedIntervals( [bedfiles[x] for x in combination ] ):
                    c.found += 1
                    if is_exclusive and isContainedInOne( contig, start, end, other_bed ):
                        c.removed += 1
                        continue
                    c.output += 1
                    outf.write( "%s\t%i\t%i\n" % (contig, start, end ) )

                outf.close()
                E.info( "combination %s finished: %s" % (tag, c))
                
                options.stdout.write( "%s\t%s\t%i\n" % (
                        ":".join( [tags[x] for x in combination] ), 
                        ":".join( [tags[x] for x in other] ), 
                        c.output ) )

    elif options.method == "unmerged-combinations":
        options.stdout.write( "track\tcombination\twithout\tcounts\n" )

        for foreground in indices:

            start = 0

            background = [x for x in indices if x != foreground ]            
            for ncombinants in range( 0, len(background) + 1 ):
                for combination in itertools.combinations( background, ncombinants ):
                    other = [ x for x in background if x not in combination ]
                    combination_bed = [ bedfiles[x] for x in combination ] 
                    other_bed = [ bedfiles[x] for x in other ]
                    tag = ":".join( [tags[foreground]] + [ tags[x] for x in combination ] )

                    E.debug( "fg=%i, combination=%s, other=%s" % (foreground, combination, other) )
                    E.debug( "combination %s started" % tag)
                    E.debug( "other: %s" % ":".join( [tags[x] for x in other]) )

                    outf = IOTools.openFile( E.getOutputFile( tag ), "w", create_dir = True)
                    c = E.Counter()
                    for bed in combineUnmergedIntervals(
                        bedfiles[foreground],
                        combination_bed ):
                        c.found += 1
                        if is_exclusive and isContainedInOne( bed.contig, bed.start, bed.end, other_bed ):
                            c.removed += 1
                            continue
                        c.output += 1
                        outf.write( "%s\n" % str(bed))

                    outf.close()
                    E.info( "combination %s finished: %s" % (tag, c))

                    options.stdout.write( "%s\t%s\t%s\t%i\n" % (
                            tags[foreground],
                            ":".join( [tags[x] for x in combination] ), 
                            ":".join( [tags[x] for x in other] ), 
                            c.output ) )


    E.Stop()


if __name__ == "__main__":
    sys.exit( main( sys.argv) )
